#!/usr/bin/env python
# proxy_receiver.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (proxy_receiver.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: proxy_receiver.

.. role:: red

BitDust proxy_receiver(at_startup) Automat

.. raw:: html

    <i>generated using <a href="https://bitdust.io/visio2python/" target="_blank">visio2python</a> tool</i><br>
    <a href="proxy_receiver.png" target="_blank">
    <img src="proxy_receiver.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`ack-received`
    * :red:`ack-timeout`
    * :red:`fail-received`
    * :red:`found-one-node`
    * :red:`inbox-packet`
    * :red:`init`
    * :red:`nodes-not-found`
    * :red:`request-timeout`
    * :red:`router-disconnected`
    * :red:`router-id-received`
    * :red:`sending-failed`
    * :red:`service-accepted`
    * :red:`service-refused`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`timer-10sec`
    * :red:`timer-15sec`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import BytesIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

import re
import time
import random

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import packetid
from bitdust.lib import strng
from bitdust.lib import serialization
from bitdust.lib import net_misc

from bitdust.automats import automat

from bitdust.main import config
from bitdust.main import settings

from bitdust.crypt import key
from bitdust.crypt import signed
from bitdust.crypt import encrypted

from bitdust.p2p import commands
from bitdust.p2p import lookup
from bitdust.p2p import online_status

from bitdust.contacts import identitycache

from bitdust.services import driver

from bitdust.transport import callback
from bitdust.transport import packet_in
from bitdust.transport import packet_out

from bitdust.transport import gateway
from bitdust.transport.proxy import proxy_interface

from bitdust.userid import identity
from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ProxyReceiver = None

#------------------------------------------------------------------------------


def GetRouterIDURL():
    global _ProxyReceiver
    if not _ProxyReceiver:
        return None
    return _ProxyReceiver.router_idurl


def GetPossibleRouterIDURL():
    global _ProxyReceiver
    if not _ProxyReceiver:
        return None
    return _ProxyReceiver.possible_router_idurl


def GetRouterIdentity():
    global _ProxyReceiver
    if not _ProxyReceiver:
        return None
    return _ProxyReceiver.router_identity


def GetRouterProtoHost():
    global _ProxyReceiver
    if not _ProxyReceiver:
        return None
    return _ProxyReceiver.router_proto_host


def ReadMyOriginalIdentitySource():
    return config.conf().getData('services/proxy-transport/my-original-identity').strip()


def WriteMyOriginalIdentitySource(new_identity_xml_src):
    return config.conf().setData('services/proxy-transport/my-original-identity', new_identity_xml_src)


def ReadCurrentRouter():
    return config.conf().getString('services/proxy-transport/current-router').strip()


def VerifyExistingRouter():
    if ReadCurrentRouter() and not ReadMyOriginalIdentitySource():
        lg.err('current router is set, but my original identity is empty')
        return False
    if not ReadCurrentRouter() and ReadMyOriginalIdentitySource():
        lg.err('current router is not set, but some wrong data found as original identity')
        return False
    return True


def LatestPacketReceived():
    global _ProxyReceiver
    if not _ProxyReceiver:
        return 0
    return _ProxyReceiver.latest_packet_received


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with proxy_receiver machine.
    """
    global _ProxyReceiver
    if event is None and not args:
        return _ProxyReceiver
    if _ProxyReceiver is None:
        # set automat name and starting state here
        _ProxyReceiver = ProxyReceiver()
    if event is not None:
        _ProxyReceiver.automat(event, *args, **kwargs)
    return _ProxyReceiver


#------------------------------------------------------------------------------


class ProxyReceiver(automat.Automat):

    """
    This class implements all the functionality of the ``proxy_receiver()``
    state machine.
    """

    timers = {
        'timer-10sec': (10.0, ['LISTEN']),
        'timer-15sec': (15.0, ['ACK?']),
        'timer-30sec': (30.0, ['FIND_NODE?']),
    }

    def __init__(self):
        """
        Builds `proxy_receiver()` state machine.
        """
        self.possible_router_idurl = None
        self.router_idurl = None
        self.router_identity = None
        self.router_id = ''
        self.router_proto_host = None
        self.request_service_packet_id = []
        self.latest_packet_received = 0
        self.router_connection_info = None
        self.traffic_in = 0
        super(ProxyReceiver, self).__init__(
            name='proxy_receiver',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
            publish_events=False,
        )

    def __repr__(self):
        return '%s_%s_%s(%s)' % (self.id, self.router_id, (self.router_connection_info or {}).get('repr', '?'), self.state)

    def to_json(self):
        j = super().to_json()
        j.update(
            {
                'proto': self.router_proto_host[0] if self.router_proto_host else '',
                'host': net_misc.pack_address(self.router_proto_host[1]) if self.router_proto_host else '',
                'idurl': self.router_idurl,
                'bytes_received': self.traffic_in,
                'bytes_sent': 0,
                'connection_info': self.router_connection_info,
                'idle_seconds': int(time.time() - self.latest_packet_received),
            }
        )
        return j

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when proxy_receiver() state were changed.
        """
        from bitdust.transport.proxy import proxy_sender
        proxy_sender.A('proxy_receiver.state', newstate)

    def A(self, event, *args, **kwargs):
        """
        The core proxy_receiver() code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFFLINE'
                self.doInit(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---ACK?---
        elif self.state == 'ACK?':
            if event == 'ack-received':
                self.state = 'SERVICE?'
                self.doSendRequestService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'OFFLINE'
                self.doNotifyFailed(*args, **kwargs)
            elif event == 'sending-failed':
                self.Retries += 1
                self.doSendMyIdentity(*args, **kwargs)
            elif (event == 'sending-failed' and self.Retries > 3) or event == 'ack-timeout' or event == 'fail-received' or event == 'timer-15sec':
                self.state = 'FIND_NODE?'
                self.doLookupRandomNode(*args, **kwargs)
        #---LISTEN---
        elif self.state == 'LISTEN':
            if event == 'router-id-received':
                self.doUpdateRouterID(*args, **kwargs)
            elif event == 'inbox-packet':
                self.doProcessInboxPacket(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopListening(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'OFFLINE'
                self.doSendCancelService(*args, **kwargs)
                self.doStopListening(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
            elif event == 'timer-10sec':
                self.doCheckPingRouter(*args, **kwargs)
            elif event == 'service-refused' or event == 'router-disconnected':
                self.state = 'FIND_NODE?'
                self.doStopListening(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
                self.doLookupRandomNode(*args, **kwargs)
        #---FIND_NODE?---
        elif self.state == 'FIND_NODE?':
            if event == 'found-one-node':
                self.state = 'ACK?'
                self.doRememberNode(*args, **kwargs)
                self.Retries = 0
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop' or event == 'nodes-not-found' or event == 'timer-30sec':
                self.state = 'OFFLINE'
                self.doNotifyFailed(*args, **kwargs)
        #---SERVICE?---
        elif self.state == 'SERVICE?':
            if event == 'service-accepted':
                self.state = 'LISTEN'
                self.doStartListening(*args, **kwargs)
                self.doNotifyConnected(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'OFFLINE'
                self.doSendCancelService(*args, **kwargs)
                self.doNotifyFailed(*args, **kwargs)
            elif event == 'request-timeout' or event == 'service-refused':
                self.state = 'FIND_NODE?'
                self.doLookupRandomNode(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'start' and self.isCurrentRouterExist(*args, **kwargs):
                self.state = 'ACK?'
                self.doLoadRouterInfo(*args, **kwargs)
                self.Retries = 0
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'start' and not self.isCurrentRouterExist(*args, **kwargs):
                self.state = 'FIND_NODE?'
                self.doLookupRandomNode(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        return None

    def isCurrentRouterExist(self, *args, **kwargs):
        """
        Condition method.
        """
        if not ReadCurrentRouter():
            return False
        if not ReadMyOriginalIdentitySource():
            return False
        return True

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        global _PacketLogFileEnabled
        _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')
        callback.add_queue_item_status_callback(self._on_queue_item_status_changed)

    def doLoadRouterInfo(self, *args, **kwargs):
        """
        Action method.
        """
        s = config.conf().getString('services/proxy-transport/current-router').strip()
        try:
            self.router_idurl = id_url.field(s.split(' ')[0])
            self.router_id = global_id.idurl2glob(self.router_idurl)
        except:
            lg.exc()
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver.doLoadRouterInfo : %s' % self.router_idurl)

    def doLookupRandomNode(self, *args, **kwargs):
        """
        Action method.
        """
        self._find_random_node(attempts=5)

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        reactor.callLater(0, self._do_send_identity_to_router, my_id.getLocalIdentity().serialize(), failed_event='fail-received')  # @UndefinedVariable
        identity_source = config.conf().getData('services/proxy-transport/my-original-identity').strip()
        if identity_source:
            if _Debug:
                lg.out(_DebugLevel, '    also sending identity loaded from "my-original-identity" config')
            reactor.callLater(0, self._do_send_identity_to_router, identity_source, failed_event='fail-received')  # @UndefinedVariable

    def doRememberNode(self, *args, **kwargs):
        """
        Action method.
        """
        self.possible_router_idurl = None
        self.router_idurl = id_url.field(args[0])
        self.router_id = global_id.idurl2glob(self.router_idurl)
        self.router_identity = None
        self.router_proto_host = None
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver.doRememberNode %r' % self.router_idurl)

    def doSendRequestService(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_send_request_service(args[0])

    def doSendCancelService(self, *args, **kwargs):
        """
        Action method.
        """
        newpacket = signed.Packet(
            commands.CancelService(),
            my_id.getIDURL(),
            my_id.getIDURL(),
            packetid.UniqueID(),
            serialization.DictToBytes({
                'name': 'service_proxy_server',
            }),
            self.router_idurl,
        )
        packet_out.create(
            newpacket,
            wide=True,
            callbacks={},
            # callbacks={
            #     commands.Ack(): self._on_request_service_ack,
            #     commands.Fail(): self._on_request_service_fail,
            # },
            response_timeout=settings.P2PTimeOut(),
        )

    def doProcessInboxPacket(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_process_inbox_packet(args[0])

    def doStartListening(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            _, info, active_router_session_machine = args[0]
            self.router_proto_host = (info.proto, info.host)
        except:
            try:
                # TODO: move that setting to separate file
                s = config.conf().getString('services/proxy-transport/current-router').strip()
                _, router_proto, router_host = s.split(' ')
                self.router_proto_host = (router_proto, strng.to_bin(router_host))
            except:
                lg.exc()
        self.router_identity = identitycache.FromCache(self.router_idurl)
        if _Debug:
            lg.args(_DebugLevel, router_idurl=self.router_idurl)
        config.conf().setString('services/proxy-transport/current-router', '%s %s %s' % (
            strng.to_text(self.router_idurl),
            strng.to_text(self.router_proto_host[0]),
            strng.to_text(self.router_proto_host[1]),
        ))
        current_identity = my_id.getLocalIdentity().serialize(as_text=True)
        previous_identity = ReadMyOriginalIdentitySource()
        if previous_identity:
            lg.err('my original identity is not empty, SKIP overwriting')
            if _Debug:
                lg.out(_DebugLevel, 'PREVIOUS ORIGINAL IDENTITY:\n%s\n' % previous_identity)
        WriteMyOriginalIdentitySource(current_identity)
        lg.info('copy of my current identity was stored as my "original" identity')
        self.request_service_packet_id = []
        callback.insert_inbox_callback(0, self._on_inbox_packet_received)
        if online_status.isKnown(self.router_idurl):
            online_status.add_online_status_listener_callback(
                idurl=self.router_idurl,
                callback_method=self._on_router_contact_status_connected,
                newstate='CONNECTED',
            )
            online_status.add_online_status_listener_callback(
                idurl=self.router_idurl,
                callback_method=self._on_router_contact_status_offline,
                newstate='OFFLINE',
            )
        active_router_session_machine.addStateChangedCallback(self._on_router_session_disconnected, oldstate='CONNECTED')
        lg.info('router %s is connected at %s://%s' % (self.router_idurl, self.router_proto_host[0], self.router_proto_host[1]))
        my_id.rebuildLocalIdentity()

    def doStopListening(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, router_idurl=self.router_idurl)
        if online_status.isKnown(self.router_idurl):
            online_status.remove_online_status_listener_callback(
                idurl=self.router_idurl,
                callback_method=self._on_router_contact_status_connected,
            )
            online_status.remove_online_status_listener_callback(
                idurl=self.router_idurl,
                callback_method=self._on_router_contact_status_offline,
            )
        active_router_session_machine_index = None
        if self.router_connection_info:
            active_router_session_machine = None
            active_router_session_machine_index = self.router_connection_info.get('index', None)
            if active_router_session_machine_index is not None:
                active_router_session_machine = automat.by_index(active_router_session_machine_index)
            if not active_router_session_machine:
                active_router_sessions = gateway.find_active_session(
                    proto=self.router_connection_info.get('proto'),
                    host=self.router_connection_info.get('host'),
                )
                if not active_router_sessions:
                    active_router_sessions = gateway.find_active_session(
                        proto=self.router_connection_info.get('proto'),
                        idurl=id_url.to_bin(self.router_idurl),
                    )
                if active_router_sessions:
                    active_router_session_machine = automat.by_index(active_router_sessions[0].index)
            if active_router_session_machine is not None:
                active_router_session_machine.removeStateChangedCallback(self._on_router_session_disconnected)
                lg.info('removed callback from router active session: %r' % active_router_session_machine)
            else:
                lg.err('did not found active router session state machine with index %s' % active_router_session_machine_index)
        WriteMyOriginalIdentitySource('')
        config.conf().setString('services/proxy-transport/current-router', '')
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        self.router_identity = None
        self.router_idurl = None
        self.router_id = ''
        self.router_proto_host = None
        self.request_service_packet_id = []
        self.router_connection_info = None
        my_id.rebuildLocalIdentity()

    def doUpdateRouterID(self, *args, **kwargs):
        """
        Action method.
        """
        newpacket, _ = args[0]
        newxml = newpacket.Payload
        newidentity = identity.identity(xmlsrc=newxml)
        cachedidentity = identitycache.FromCache(self.router_idurl)
        if self.router_idurl != newidentity.getIDURL():
            lg.warn('router idurl is unrecognized from response %r != %r' % (self.router_idurl, newidentity.getIDURL()))
            return
        if newidentity.serialize() != cachedidentity.serialize():
            lg.warn('cached identity is not same, router identity changed')
        self.router_identity = newidentity

    def doCheckPingRouter(self, *args, **kwargs):
        """
        Action method.
        """
        if self.router_connection_info:
            gateway.send_keep_alive(self.router_connection_info['proto'], self.router_connection_info['host'])
        live_time = time.time() - self.latest_packet_received
        if live_time < 60.0:
            if _Debug:
                lg.out(_DebugLevel, 'proxy_receiver.doCheckPingRouter OK, latest packet received %f sec ago' % live_time)
            return
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver.doCheckPingRouter to %s' % self.router_idurl)
        identity_source = config.conf().getData('services/proxy-transport/my-original-identity').strip()
        if identity_source:
            if _Debug:
                lg.out(_DebugLevel, '    "my-original-identity" prepared for sending')
        else:
            identity_source = my_id.getLocalIdentity().serialize()
            if _Debug:
                lg.out(_DebugLevel, '    local identity prepared for sending')
        self._do_send_identity_to_router(identity_source, failed_event='router-disconnected')

    def doNotifyConnected(self, *args, **kwargs):
        """
        Action method.
        """
        proxy_interface.interface_receiving_started(self.router_idurl, {
            'router_idurl': self.router_idurl,
        })

    def doNotifyDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        proxy_interface.interface_disconnected().addErrback(lambda _: None)

    def doNotifyFailed(self, *args, **kwargs):
        """
        Action method.
        """
        proxy_interface.interface_receiving_failed()

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        global _PacketLogFileEnabled
        global _ProxyReceiver
        _PacketLogFileEnabled = False
        callback.remove_queue_item_status_callback(self._on_queue_item_status_changed)
        self.possible_router_idurl = None
        self.router_idurl = None
        self.router_id = ''
        self.router_identity = None
        self.router_proto_host = None
        self.request_service_packet_id = []
        self.latest_packet_received = 0
        self.router_connection_info = None
        self.traffic_in = 0
        self.destroy()
        del _ProxyReceiver
        _ProxyReceiver = None

    def _do_process_inbox_packet(self, *args, **kwargs):
        newpacket, info, _, _ = args[0]
        block = encrypted.Unserialize(newpacket.Payload)
        if block is None:
            lg.err('reading data from %s' % newpacket.CreatorID)
            return
        try:
            session_key = key.DecryptLocalPrivateKey(block.EncryptedSessionKey)
            padded_data = key.DecryptWithSessionKey(session_key, block.EncryptedData, session_key_type=block.SessionKeyType)
            inpt = BytesIO(padded_data[:int(block.Length)])
            data = inpt.read()
        except:
            lg.err('reading data from %s' % newpacket.CreatorID)
            lg.exc()
            try:
                inpt.close()
            except:
                pass
            return
        inpt.close()

        if newpacket.Command == commands.RelayAck():
            try:
                ack_info = serialization.BytesToDict(data, keys_to_text=True, values_to_text=True)
            except:
                lg.exc()
                return
            if _Debug:
                lg.out(_DebugLevel, '<<<Relay-ACK %s:%s from %s://%s with %d bytes %s' % (ack_info['command'], ack_info['packet_id'], info.proto, info.host, len(data), ack_info['error']))
            if _PacketLogFileEnabled:
                lg.out(
                    0, '                \033[0;49;33mRELAY ACK %s(%s) with %d bytes from %s to %s TID:%s\033[0m' %
                    (ack_info['command'], ack_info['packet_id'], info.bytes_received, global_id.UrlToGlobalID(ack_info['from']), global_id.UrlToGlobalID(ack_info['to']), info.transfer_id), log_name='packet', showtime=True
                )
            from bitdust.transport.proxy import proxy_sender
            if proxy_sender.A():
                proxy_sender.A('relay-ack', ack_info, info)
            return True

        if newpacket.Command == commands.RelayFail():
            try:
                fail_info = serialization.BytesToDict(data, keys_to_text=True, values_to_text=True)
            except:
                lg.exc()
                return
            if _Debug:
                lg.out(_DebugLevel, '<<<Relay-FAIL %s:%s from %s://%s with %d bytes %s' % (fail_info['command'], fail_info['packet_id'], info.proto, info.host, len(data), fail_info['error']))
            if _PacketLogFileEnabled:
                lg.out(
                    0, '                \033[0;49;33mRELAY FAIL %s(%s) with %d bytes from %s to %s TID:%s\033[0m' %
                    (fail_info['command'], fail_info['packet_id'], info.bytes_received, global_id.UrlToGlobalID(fail_info['from']), global_id.UrlToGlobalID(fail_info['to']), info.transfer_id), log_name='packet', showtime=True
                )
            from bitdust.transport.proxy import proxy_sender
            if proxy_sender.A():
                proxy_sender.A('relay-failed', fail_info, info)
            return True

        routed_packet = signed.Unserialize(data)
        if not routed_packet:
            lg.err('unserialize packet failed from %s' % newpacket.CreatorID)
            return

        if _Debug:
            lg.out(_DebugLevel, '<<<Relay-IN %s from %s://%s with %d bytes' % (str(routed_packet), info.proto, info.host, len(data)))
        if _PacketLogFileEnabled:
            lg.out(
                0, '                \033[0;49;33mRELAY IN %s(%s) with %d bytes from %s to %s TID:%s\033[0m' %
                (routed_packet.Command, routed_packet.PacketID, info.bytes_received, global_id.UrlToGlobalID(info.sender_idurl), global_id.UrlToGlobalID(routed_packet.RemoteID), info.transfer_id), log_name='packet', showtime=True
            )

        if routed_packet.Command == commands.Identity():
            if _Debug:
                lg.out(_DebugLevel, '    found identity in relay packet %s' % routed_packet)
            newidentity = identity.identity(xmlsrc=routed_packet.Payload)
            idurl = newidentity.getIDURL()
            if not identitycache.HasKey(idurl):
                lg.info('received new identity %s rev %r' % (idurl.original(), newidentity.getRevisionValue()))
            if not identitycache.UpdateAfterChecking(idurl, routed_packet.Payload):
                lg.warn('ERROR has non-Valid identity')
                return

        if routed_packet.Command in [
            commands.Relay(),
            commands.RelayIn(),
        ] and routed_packet.PacketID.lower().startswith('identity:'):
            if _Debug:
                lg.out(_DebugLevel, '    found routed identity in relay packet %s' % routed_packet)
            try:
                routed_identity = signed.Unserialize(routed_packet.Payload)
                newidentity = identity.identity(xmlsrc=routed_identity.Payload)
                idurl = newidentity.getIDURL()
                if not identitycache.HasKey(idurl):
                    lg.warn('received new "routed" identity: %s' % idurl)
                if not identitycache.UpdateAfterChecking(idurl, routed_identity.Payload):
                    lg.warn('ERROR has non-Valid identity')
                    return
            except:
                lg.exc()

        if newpacket.Command == commands.RelayIn() and routed_packet.Command == commands.Fail():
            if routed_packet.Payload == b'route not exist' or routed_packet.Payload == b'route already closed':
                for pout in packet_out.search_by_packet_id(routed_packet.PacketID):
                    if _Debug:
                        lg.dbg(_DebugLevel, 'received %r from %r, outgoing packet is failed: %r' % (routed_packet.Payload, newpacket.CreatorID, pout))
                    pout.automat('request-failed')
                return

        self.traffic_in += len(data)
        packet_in.process(routed_packet, info)
        del block
        del data
        del padded_data
        del inpt
        del session_key
        del routed_packet

    def _do_send_identity_to_router(self, identity_source, failed_event):
        try:
            identity_obj = identity.identity(xmlsrc=identity_source)
        except:
            lg.exc()
            return
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver._do_send_identity_to_router to %s' % self.router_idurl)
            lg.out(_DebugLevel, '        contacts=%r, sources=%r' % (identity_obj.contacts, identity_obj.getSources(as_originals=True)))
        newpacket = signed.Packet(
            Command=commands.Identity(),
            OwnerID=my_id.getIDURL(),
            CreatorID=my_id.getIDURL(),
            PacketID=('proxy_receiver:%s' % packetid.UniqueID()),
            Payload=identity_obj.serialize(),
            RemoteID=self.router_idurl,
        )
        packet_out.create(
            newpacket,
            wide=True,
            callbacks={
                commands.Ack(): lambda response, info: self.automat('ack-received', (response, info)),
                commands.Fail(): lambda x: self.automat(failed_event),
                None: lambda pkt_out: self.automat('ack-timeout', pkt_out),
                'failed': lambda pkt_out, error_message: self.automat('sending-failed', (pkt_out, error_message)),
            },
            keep_alive=True,
            response_timeout=settings.P2PTimeOut(),
        )

    def _do_send_request_service(self, *args, **kwargs):
        if len(self.request_service_packet_id) >= 10:
            if _Debug:
                lg.warn('too many service requests to %r' % self.router_idurl)
            self.automat('service-refused', *args, **kwargs)
            return
        orig_identity = config.conf().getData('services/proxy-transport/my-original-identity').strip()
        if not orig_identity:
            orig_identity = my_id.getLocalIdentity().serialize(as_text=True)
        service_info = {
            'name': 'service_proxy_server',
            'payload': {
                'identity': orig_identity,
            },
        }
        newpacket = signed.Packet(
            commands.RequestService(),
            my_id.getIDURL(),
            my_id.getIDURL(),
            packetid.UniqueID(),
            serialization.DictToBytes(service_info, values_to_text=True),
            self.router_idurl,
        )
        packet_out.create(
            newpacket,
            wide=False,
            callbacks={
                commands.Ack(): self._on_request_service_ack,
                commands.Fail(): self._on_request_service_fail,
                # 'timeout': lambda pkt_out, err: self.automat('request-timeout', pkt_out),
                None: lambda pkt_out: self.automat('request-timeout', pkt_out),
            },
            response_timeout=settings.P2PTimeOut(),
        )
        self.request_service_packet_id.append(newpacket.PacketID)

    def _on_nodes_lookup_finished(self, idurls, attempts):
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver._on_nodes_lookup_finished : %r' % idurls)
        for idurl in idurls:
            ident = identitycache.FromCache(idurl)
            remoteprotos = set(ident.getProtoOrder())
            myprotos = set(my_id.getLocalIdentity().getProtoOrder())
            if len(myprotos.intersection(remoteprotos)) > 0:
                self.possible_router_idurl = id_url.field(idurl)
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_receiver._on_nodes_lookup_finished found : %r' % self.possible_router_idurl)
                self.automat('found-one-node', self.possible_router_idurl)
                return
        if attempts > 0:
            self._find_random_node(attempts - 1)
        else:
            self.automat('nodes-not-found')

    def _find_random_node(self, attempts):
        preferred_routers = []
        preferred_routers_raw = config.conf().getString('services/proxy-transport/preferred-routers').strip()
        if preferred_routers_raw:
            preferred_routers_list = re.split('\n|,|;| ', preferred_routers_raw)
            preferred_routers.extend(preferred_routers_list)
        if preferred_routers:
            self.possible_router_idurl = id_url.field(random.choice(preferred_routers))
            if _Debug:
                lg.out(_DebugLevel, 'proxy_receiver._find_random_node selected random item from preferred_routers: %r' % self.possible_router_idurl)
            idcache_defer = identitycache.immediatelyCaching(self.possible_router_idurl)
            idcache_defer.addCallback(lambda *args: self.automat('found-one-node', self.possible_router_idurl))
            idcache_defer.addErrback(lambda err: self.automat('nodes-not-found') and None)
            return
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver._find_random_node will start DHT lookup')
        tsk = lookup.random_proxy_router()
        if tsk:
            tsk.result_defer.addCallback(self._on_nodes_lookup_finished, attempts=attempts)
            tsk.result_defer.addErrback(lambda err: self.automat('nodes-not-found'))
        else:
            self.automat('nodes-not-found')

    def _on_request_service_ack(self, response, info):
        self.router_connection_info = None
        if response.PacketID not in self.request_service_packet_id:
            lg.warn('wrong PacketID in response: %s, but outgoing was : %s' % (response.PacketID, str(self.request_service_packet_id)))
            self.automat('service-refused', (response, info))
            return
        if response.PacketID in self.request_service_packet_id:
            self.request_service_packet_id.remove(response.PacketID)
        else:
            lg.warn('%s was not found in pending requests: %s' % (response.PacketID, self.request_service_packet_id))
        if _Debug:
            lg.out(_DebugLevel, 'proxy_receiver._on_request_service_ack : %s' % str(response.Payload))
        if self.router_idurl != response.CreatorID:
            lg.err('received unexpected response from another node: %r ~ %r' % (self.router_idurl, response.CreatorID))
            self.automat('service-refused', (response, info))
            return
        service_ack_info = strng.to_text(response.Payload)
        if service_ack_info.startswith('rejected'):
            self.automat('service-refused', (response, info))
            return
        active_router_sessions = gateway.find_active_session(info.proto, host=info.host)
        if not active_router_sessions:
            active_router_sessions = gateway.find_active_session(info.proto, idurl=id_url.to_bin(response.CreatorID))
        if not active_router_sessions:
            lg.err('active connection with proxy router at %s:%s was not found' % (info.proto, info.host))
            if _Debug:
                lg.args(_DebugLevel, router_idurl=self.router_idurl, ack_packet=info, active_sessions=gateway.list_active_sessions(info.proto))
            self.automat('service-refused', (response, info))
            return
        self.router_connection_info = {
            'id': active_router_sessions[0].id,
            'index': active_router_sessions[0].index,
            'repr': repr(active_router_sessions[0]),
            'proto': info.proto,
            'host': info.host,
            'idurl': self.router_idurl,
            'global_id': global_id.UrlToGlobalID(self.router_idurl),
        }
        active_router_session_machine = automat.by_index(self.router_connection_info['index'])
        if active_router_session_machine is None:
            lg.err('did not found proxy router session state machine instance: %s' % self.router_connection_info)
            self.router_connection_info = None
            if _Debug:
                lg.args(_DebugLevel, automats=automat.objects())
            self.automat('service-refused', (response, info))
            return
        lg.info('found active session for proxy router: %s' % active_router_session_machine)
        self.automat('service-accepted', (response, info, active_router_session_machine))

    def _on_request_service_fail(self, response, info):
        if response.PacketID not in self.request_service_packet_id:
            lg.warn('wrong PacketID in response: %s, but outgoing was : %s' % (response.PacketID, str(self.request_service_packet_id)))
        else:
            self.request_service_packet_id.remove(response.PacketID)
        self.automat('service-refused', (response, info))

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        if newpacket.Command == commands.Identity() and \
                newpacket.CreatorID == self.router_idurl and \
                newpacket.RemoteID == my_id.getIDURL():
            self.automat('router-id-received', (newpacket, info))
            self.latest_packet_received = time.time()
            return True
        if newpacket.CreatorID == self.router_idurl:
            self.latest_packet_received = time.time()
        if newpacket.Command in [
            commands.Relay(),
            commands.RelayIn(),
            commands.RelayAck(),
            commands.RelayFail(),
        ]:
            if driver.is_enabled('service_proxy_server'):
                # TODO:
                # in case this node already running proxy router service this will not work
                # actually you can not use proxy transport for receiving and running proxy router at same time
                # need to change one of the services to solve that dependency and prevent this
                return False
            self.automat('inbox-packet', (newpacket, info, status, error_message))
            return True
        return False

    def _on_router_contact_status_connected(self, oldstate, newstate, event_string, *args, **kwargs):
        lg.info('router %r contact status online: %s->%s after "%s"' % (self.router_idurl, oldstate, newstate, event_string))

    def _on_router_contact_status_offline(self, oldstate, newstate, event_string, *args, **kwargs):
        lg.warn('router %r contact status offline: %s->%s after "%s"' % (self.router_idurl, oldstate, newstate, event_string))
        # self.automat('router-disconnected')

    def _on_router_session_disconnected(self, oldstate, newstate, event_string, *args, **kwargs):
        lg.err('router session disconnected: %s->%s because of %r' % (oldstate, newstate, event_string))
        self.automat('router-disconnected')

    def _on_queue_item_status_changed(self, pkt_out, status, error=''):
        from bitdust.transport.proxy import proxy_receiver
        if status == 'finished':
            return False
        if error != 'connection failed':
            return False
        if not pkt_out.remote_idurl or not pkt_out.outpacket:
            return False
        if id_url.to_bin(pkt_out.remote_idurl) == pkt_out.outpacket.RemoteID.to_bin():
            return False
        if not proxy_receiver.GetRouterIDURL():
            return False
        if pkt_out.remote_idurl != proxy_receiver.GetRouterIDURL():
            return False
        lg.err('connection failed with current proxy router, must reconnect to another router: %r %r %r' % (pkt_out, status, error))
        self.automat('router-disconnected')
        return True


#------------------------------------------------------------------------------


def main():
    from twisted.internet import reactor  # @UnresolvedImport
    reactor.callWhenRunning(A, 'init')  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    main()



#!/usr/bin/python
# identitycache.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (identitycache.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: identitycache.

Here we store a local copies of identities. This fetches identities off
the web and stores an XML copy in file and an identity object in a
dictionary. Other parts of BitDust call this to get an identity using an
IDURL. So this is a local cache of user ID's.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys
import time

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred, DeferredList, CancelledError

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import net_misc
from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.userid import identity
from bitdust.userid import id_url

from bitdust.contacts import identitydb

from bitdust.p2p import p2p_stats

#------------------------------------------------------------------------------

_CachingTasks = {}
_LastTimeCached = {}
_OverriddenIdentities = {}

#-------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'identitycache.init')
    identitydb.clear()


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'identitycache.shutdown')


#------------------------------------------------------------------------------


def caching():
    global _CachingTasks
    return _CachingTasks


def add_callback(idurl, ignore_errors=False):
    idurl = id_url.to_original(idurl)
    defer_obj = Deferred()
    defer_obj.addErrback(on_caching_task_failed, idurl, ignore_errors)
    if idurl not in caching():
        caching()[idurl] = []
    caching()[idurl].append(defer_obj)
    return defer_obj


#------------------------------------------------------------------------------


def get_one(idurl):
    """
    An alias for `FromCache()` method.
    """
    return FromCache(id_url.to_original(idurl))


def start_one(idurl, timeout=15, try_other_sources=True):
    """
    An alias for `immediatelyCaching()` method.
    """
    return immediatelyCaching(id_url.to_original(idurl), timeout=timeout, try_other_sources=try_other_sources)


def start_multiple(idurl_list, timeout=15, try_other_sources=True):
    """
    Executes in parallel multiple `immediatelyCaching()` calls on the list of IDURL items.
    Returns `DeferredList` object.
    """
    idurl_list = list(set(id_url.to_original_list(idurl_list)))
    dl = []
    for idurl in idurl_list:
        if idurl:
            dl.append(immediatelyCaching(idurl, timeout=timeout, try_other_sources=try_other_sources))
    return DeferredList(dl, consumeErrors=True)


#------------------------------------------------------------------------------


def Clear(excludeList=None):
    """
    Clear all cached identities.
    """
    identitydb.clear(excludeList)


def CacheLen():
    """
    Return a number of items in the cache.
    """
    return identitydb.size()


def PrintID(idurl):
    """
    For debug, print an item from cache.
    """
    identitydb.print_id(idurl)


def PrintCacheKeys():
    """
    For debug, print all items keys in the cache.
    """
    identitydb.print_keys()


def PrintAllInCache():
    """
    For debug, print completely all cache.
    """
    identitydb.print_cache()


def Items():
    return identitydb.cache()


def HasKey(idurl):
    """
    Check for some user IDURL in the cache.
    """
    return identitydb.has_idurl(idurl) or IsOverridden(idurl)


def GetLastModifiedTime(idurl):
    return identitydb.get_last_modified_time(idurl)


def HasFile(idurl):
    return identitydb.has_file(idurl)


def FromCache(idurl):
    """
    Get identity object from cache.
    """
    if IsOverridden(idurl):
        overridden_xmlsrc = ReadOverriddenIdentityXMLSource(idurl)
        if overridden_xmlsrc:
            if _Debug:
                lg.out(_DebugLevel, '        returning overridden identity (%d bytes) for %s' % (len(overridden_xmlsrc), idurl))
            return identity.identity(xmlsrc=overridden_xmlsrc)
    return identitydb.get_ident(idurl)


def GetLatest(idurl):
    """
    Returns latest copy from cache or fire `immediatelyCaching`,
    result is a `Deferred` object.
    """
    idurl = id_url.to_original(idurl)
    known = FromCache(idurl)
    result = Deferred()
    if known:
        result.callback(known)
    else:
        d = immediatelyCaching(idurl)
        d.addCallback(lambda _: result.callback(FromCache(idurl)))
        d.addErrback(lambda err: result.errback(err) and None)
    return result


def GetPublicKey(idurl):
    """
    If given `idurl` is known will return public key from the corresponding identity file otherwise `None`.
    """
    ident = FromCache(idurl)
    if not ident:
        return None
    return ident.getPublicKey()


def GetIDURLsByContact(contact):
    """
    In the ``identitydb`` code we keep track of all identity objects and
    prepare an index of all known contacts.

    So we can try to detect who is sending us a packet when got a packet
    from known contact address. This is to get a list of known ID's in
    the cache for that contact.
    """
    return identitydb.get_idurls_by_contact(contact)


def GetIDURLByIPPort(ip, port):
    """
    Same as previous method but the index is created from IP:PORT parts of
    identities contacts.
    """
    return identitydb.get_idurl_by_ip_port(ip, port)


def GetContacts(idurl):
    """
    This is another one index - return a set of contacts for given IDURL.
    Instead of read identity object and parse it every time - this method can be used.
    This is a "cached" info.
    """
    return identitydb.idcontacts(idurl)


def Remove(idurl):
    """
    Remove an item from cache.
    """
    return identitydb.remove(idurl)


def UpdateAfterChecking(idurl, xml_src):
    """
    Need to call that method to update the cache when some identity sources is
    changed.
    """
    return identitydb.update(idurl, xml_src)


def RemapContactAddress(address):
    """
    For local peers in same sub network we need to use local IP, not external
    IP. We pass local IP to transports to send packets inside sub network.

    TODO: Would be great to get rid of that - transport must keep track of local and external situations.
    So this is another index - IDURL to local IP, see identitydb.
    """
    idurl = GetIDURLByIPPort(address[0], address[1])
    if idurl is not None and HasLocalIP(idurl):
        newaddress = (GetLocalIP(idurl), address[1])
        if _Debug:
            lg.out(_DebugLevel, 'identitycache.RemapContactAddress for %s [%s] -> [%s]' % (nameurl.GetName(idurl), str(address), str(newaddress)))
        return newaddress
    return address


def OverrideIdentity(idurl, xml_src):
    """
    Used by proxy router to remember the original identity of the routing node.
    """
    global _OverriddenIdentities
    idurl = id_url.field(idurl)
    if not idurl.is_latest():
        if idurl.original() in _OverriddenIdentities:
            if idurl.to_bin() not in _OverriddenIdentities:
                _OverriddenIdentities[idurl.to_bin()] = _OverriddenIdentities.pop(idurl.original())
                lg.info('detected and processed idurl rotate for overridden identity : %r -> %r' % (idurl.original(), idurl.to_bin()))
    idurl = id_url.to_bin(idurl)
    xml_src = strng.to_text(xml_src.strip())
    if idurl in _OverriddenIdentities:
        if _OverriddenIdentities[idurl] == xml_src:
            if _Debug:
                lg.out(_DebugLevel, 'identitycache.OverrideIdentity SKIPPED "%s", no changes' % idurl)
            return False
        if _Debug:
            lg.out(_DebugLevel, 'identitycache.OverrideIdentity replacing overriden identity %r with new one' % idurl)
            lg.out(_DebugLevel, '\nOVERRIDDEN OLD:\n' + _OverriddenIdentities[idurl])
            lg.out(_DebugLevel, '\nOVERRIDDEN NEW:\n' + xml_src)
    else:
        orig = ''
        if identitydb.has_idurl(idurl):
            orig = identitydb.get_ident(idurl).serialize(as_text=True)
        if orig and orig == xml_src:
            if _Debug:
                lg.out(_DebugLevel, 'identitycache.OverrideIdentity SKIPPED %r , overridden copy is the same as original' % idurl)
            return False
        if _Debug:
            lg.out(_DebugLevel, 'identitycache.OverrideIdentity replacing original identity for %r' % idurl)
            lg.out(_DebugLevel, '\nORIGINAL:\n' + orig)
            lg.out(_DebugLevel, '\nNEW:\n' + xml_src)
    _OverriddenIdentities[idurl] = xml_src
    if _Debug:
        lg.out(_DebugLevel, '    total number of overrides: %d' % len(_OverriddenIdentities))
    return True


def StopOverridingIdentity(idurl):
    global _OverriddenIdentities
    idurl = id_url.field(idurl)
    if not idurl.is_latest():
        if idurl.original() in _OverriddenIdentities:
            if idurl.to_bin() not in _OverriddenIdentities:
                _OverriddenIdentities[idurl.to_bin()] = _OverriddenIdentities.pop(idurl.original())
                lg.info('detected and processed idurl rotate for overridden identity : %r -> %r' % (idurl.original(), idurl.to_bin()))
    idurl = id_url.to_bin(idurl)
    result = _OverriddenIdentities.pop(idurl, None)
    if _Debug:
        lg.out(_DebugLevel, 'identitycache.StopOverridingIdentity   removed overridden source for %s' % idurl)
        if result:
            lg.out(_DebugLevel, '    previous overridden identity was %d bytes' % len(result))
        lg.out(_DebugLevel, '            total number of overrides is %d' % len(_OverriddenIdentities))
    return result


def IsOverridden(idurl):
    global _OverriddenIdentities
    idurl = id_url.field(idurl)
    if not idurl.is_latest():
        if idurl.original() in _OverriddenIdentities:
            if idurl.to_bin() not in _OverriddenIdentities:
                _OverriddenIdentities[idurl.to_bin()] = _OverriddenIdentities.pop(idurl.original())
                lg.info('detected and processed idurl rotate for overridden identity : %r -> %r' % (idurl.original(), idurl.to_bin()))
    idurl = id_url.to_bin(idurl)
    return idurl in _OverriddenIdentities


def ReadOverriddenIdentityXMLSource(idurl):
    global _OverriddenIdentities
    idurl = id_url.field(idurl)
    if not idurl.is_latest():
        if idurl.original() in _OverriddenIdentities:
            if idurl.to_bin() not in _OverriddenIdentities:
                _OverriddenIdentities[idurl.to_bin()] = _OverriddenIdentities.pop(idurl.original())
                lg.info('detected and processed idurl rotate for overridden identity : %r -> %r' % (idurl.original(), idurl.to_bin()))
    idurl = id_url.to_bin(idurl)
    return _OverriddenIdentities.get(idurl, None)


#------------------------------------------------------------------------------


def getPageSuccess(src, idurl):
    """
    This is called when requested identity source gets received.
    """
    UpdateAfterChecking(idurl, src)
    p2p_stats.count_identity_cache(idurl, len(src))
    return src


def getPageFail(x, idurl):
    """
    This is called when identity request is failed.
    """
    if _Debug:
        lg.out(_DebugLevel, 'identitycache.getPageFail NETERROR in request to ' + idurl)
    p2p_stats.count_identity_cache(idurl, 0)
    return x


def pageRequestTwisted(idurl, timeout=15):
    """
    Request an HTML page - this can be an user identity.
    """
    d = net_misc.getPageTwisted(idurl, timeout)
    d.addCallback(getPageSuccess, idurl)
    d.addErrback(getPageFail, idurl)
    return d


def scheduleForCaching(idurl, timeout=15):
    """
    Even if we have a copy in cache we are to try and read another one.
    """
    if _Debug:
        lg.out(_DebugLevel, 'identitycache.scheduleForCaching %r' % idurl)
    return pageRequestTwisted(idurl, timeout)


#------------------------------------------------------------------------------


def last_time_cached(idurl):
    global _LastTimeCached
    idurl = id_url.to_original(idurl)
    if not idurl:
        return None
    return _LastTimeCached.get(idurl, None)


def on_caching_task_failed(err, idurl, ignore_errors):
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, err=err)
    if ignore_errors:
        return None
    if err.type == CancelledError:
        return None
    return err


def immediatelyCaching(idurl, timeout=15, try_other_sources=True, ignore_errors=False):
    """
    A smart method to cache some identity and get results in callbacks.
    """
    idurl = id_url.to_original(idurl)
    if not idurl:
        raise Exception('can not cache, idurl is empty')

    def _success(src, idurl):
        global _LastTimeCached
        idurl = id_url.to_original(idurl)
        defer_results = caching().pop(idurl, [])
        if _Debug:
            lg.args(_DebugLevel, src=type(src), idurl=idurl, defer_results=len(defer_results))
        if UpdateAfterChecking(idurl, src):
            for result in defer_results:
                if result and not result.called:
                    reactor.callLater(0, result.callback, src)  # @UndefinedVariable
            if _Debug:
                lg.out(_DebugLevel, '[cached] %s' % idurl)
            p2p_stats.count_identity_cache(idurl, len(src))
            _LastTimeCached[idurl] = time.time()
        else:
            for result in defer_results:
                if result and not result.called:
                    reactor.callLater(0, result.errback, Exception(src))  # @UndefinedVariable
            lg.warn('[cache error] %s is not valid' % idurl)
            p2p_stats.count_identity_cache(idurl, 0)
            _LastTimeCached.pop(idurl, None)
        defer_results = []
        del defer_results
        return src

    def _next_source(resp, idurl, sources, pos):
        global _LastTimeCached
        if _Debug:
            lg.args(_DebugLevel, resp=resp, idurl=idurl, pos=pos, sources=len(sources))

        if pos >= len(sources):
            lg.warn('[cache failed] %r and also %d other sources' % (idurl, len(sources)))
            defer_results = caching().pop(idurl, [])
            for result in defer_results:
                if result and not result.called:
                    reactor.callLater(0, result.errback, Exception('cache failed from multiple sources'))  # @UndefinedVariable
            p2p_stats.count_identity_cache(idurl, 0)
            _LastTimeCached.pop(idurl, None)
            defer_results = []
            del defer_results
            return None

        next_idurl = sources[pos]
        next_idurl = id_url.to_original(next_idurl)
        if _Debug:
            lg.args(_DebugLevel, next_idurl=next_idurl)

        d = net_misc.getPageTwisted(url=next_idurl, timeout=timeout)
        d.addCallback(_success, idurl)
        d.addErrback(_next_source, idurl, sources, pos + 1)
        return None

    def _fail(err, idurl):
        global _LastTimeCached
        idurl = id_url.to_original(idurl)
        if _Debug:
            lg.args(_DebugLevel, err=err, idurl=idurl)

        if not try_other_sources:
            p2p_stats.count_identity_cache(idurl, 0)
            _LastTimeCached.pop(idurl, None)
            lg.warn('[cache failed] %s : %s' % (idurl, err.getErrorMessage()))
            defer_results = caching().pop(idurl, [])
            for result in defer_results:
                if result and not result.called:
                    reactor.callLater(0, result.errback, err)  # @UndefinedVariable
            defer_results = []
            del defer_results
            return None

        sources = []
        latest_ident = None
        latest_idurl, _ = id_url.get_latest_revision(idurl)
        if not latest_idurl:
            latest_idurl = idurl
        if latest_idurl:
            latest_ident = identitydb.get_ident(latest_idurl)
        if latest_ident:
            sources.extend(list(latest_ident.getSources(as_originals=True)))
        if not sources:
            pub_key = id_url.known().get(latest_idurl)
            if pub_key:
                known_sources = id_url.sources(pub_key)
                for another_idurl in reversed(known_sources):
                    if another_idurl != latest_idurl and another_idurl != idurl:
                        if another_idurl not in sources:
                            sources.append(another_idurl)
        if idurl in sources:
            sources.remove(idurl)

        if _Debug:
            lg.args(_DebugLevel, idurl=idurl, latest_idurl=latest_idurl, latest_ident=latest_ident, sources=sources)

        if sources:
            lg.warn('[cache failed] %s : %s  but will try %d more sources' % (idurl, err.getErrorMessage(), len(sources)))
            _next_source(None, idurl, sources, 0)
            return None

        p2p_stats.count_identity_cache(idurl, 0)
        _LastTimeCached.pop(idurl, None)
        lg.warn('[cache failed] and also no other sources found %s : %s' % (idurl, err.getErrorMessage()))
        defer_results = caching().pop(idurl, [])
        if _Debug:
            lg.args(_DebugLevel, known=len(id_url.known().keys()), defer_results=len(defer_results))
        for result in defer_results:
            if result and not result.called:
                reactor.callLater(0, result.errback, err)  # @UndefinedVariable
        defer_results = []
        del defer_results
        return None

    def _start_one(idurl, ignore_errors):
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl)
        defer_obj = add_callback(idurl, ignore_errors=ignore_errors)
        d = net_misc.getPageTwisted(idurl, timeout)
        d.addCallback(_success, idurl)
        d.addErrback(_fail, idurl)
        return defer_obj

    if idurl in caching():
        if _Debug:
            lg.out(_DebugLevel, 'identitycache.immediatelyCaching already has tasks for %r' % idurl)
        defer_obj = add_callback(idurl, ignore_errors=ignore_errors)
        return defer_obj

    if _Debug:
        lg.out(_DebugLevel, 'identitycache.immediatelyCaching started new task for %r' % idurl)
    return _start_one(idurl, ignore_errors=ignore_errors)


#------------------------------------------------------------------------------


def SetLocalIPs(local_ips):
    """
    This method is to build an index for local IP's.
    """
    return identitydb.update_local_ips_dict(local_ips)


def GetLocalIP(idurl):
    """
    If known, return a local IP for given user IDURL.
    """
    return identitydb.get_local_ip(idurl)


def HasLocalIP(idurl):
    """
    Return True if at least one local IP is known for that IDURL.
    """
    return identitydb.has_local_ip(idurl)


def SearchLocalIP(ip):
    """
    For given IP search all users in the cache with same local IP.
    """
    return identitydb.search_local_ip(ip)


#------------------------------------------------------------------------------


def _test():
    import logging
    logging.basicConfig(level=logging.DEBUG)
    from twisted.internet import reactor  # @UnresolvedImport
    from twisted.internet.defer import setDebugging
    setDebugging(True)
    # from twisted.python import log as twisted_log
    # twisted_log.startLogging(sys.stdout)
    lg.set_debug_level(20)

    from bitdust.main import settings
    settings.init()
    settings.update_proxy_settings()

    init()

    def _resp(src):
        print(src)
        reactor.stop()  # @UndefinedVariable

    immediatelyCaching(sys.argv[1]).addBoth(_resp)
    reactor.run()  # @UndefinedVariable
    shutdown()
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    _test()



#!/usr/bin/python
# global_id.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (global_id.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: global_id.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import re

#------------------------------------------------------------------------------

from bitdust.lib import strng

#------------------------------------------------------------------------------

_FORMAT_GLOBAL_ID = '{key_alias}${username}@{host}'
_FORMAT_GLOBAL_ID_USER = '{username}@{host}'
_FORMAT_GLOBAL_ID_USER_KEY = '{user}!{key_alias}'
_FORMAT_GLOBAL_ID_KEY_USER = '{key_alias}${user}'
_FORMAT_GLOBAL_ID_QUEUE_ID = '{queue_alias}&{owner_id}&{supplier_id}'
_FORMAT_GLOBAL_CUSTOMER_QUEUE_ID = '{queue_alias}&{customer_id}&{position}'

_REGEX_GLOBAL_ID_USER_KEY = '^(?P<user>[a-z0-9-_]+)\!(?P<key_alias>[a-z0-9-_]+)$'
_REGEX_GLOBAL_ID_KEY_USER = '^(?P<key_alias>[a-z0-9-_]+)\$(?P<user>[a-z0-9-_]+)$'
_REGEX_GLOBAL_ID_QUEUE_ID = '^(?P<queue_alias>[a-z0-9-_]+)\&(?P<owner_id>[a-z0-9-_\@\.]+)\&(?P<supplier_id>[a-z0-9-_\@\.]+)$'
_REGEX_GLOBAL_CUSTOMER_QUEUE_ID = '^(?P<queue_alias>[a-z0-9-_]+)\&(?P<customer_id>[a-z0-9-_\@\.]+)\&(?P<position>[0-9]+)$'

#------------------------------------------------------------------------------

_REGEX_OBJ_GLOBAL_ID_QUEUE_ID = None

#------------------------------------------------------------------------------


def idurl2glob(idurl):
    """
    Alias.
    """
    return UrlToGlobalID(idurl)


def glob2idurl(glob_id, as_field=True):
    """
    Alias.
    """
    return GlobalUserToIDURL(glob_id, as_field=as_field)


def latest_glob_id(glob_id):
    return glob2idurl(glob_id, as_field=True).to_id()


#------------------------------------------------------------------------------


def MakeGlobalKeyID(key_alias, user_id):
    return _FORMAT_GLOBAL_ID_KEY_USER.format(
        key_alias=key_alias,
        user=user_id,
    )


#------------------------------------------------------------------------------


def MakeGlobalID(
    idurl=None,
    user=None,
    idhost=None,
    customer=None,
    key_alias=None,
    path=None,
    version=None,
    key_id=None,
):
    """
    Based on input parameters returns string like this:

        group_abc$alice@first-machine.com:animals/cat.png#F20160313043757PM

    """
    output_format = _FORMAT_GLOBAL_ID_KEY_USER
    out = ''
    if key_id:
        out = key_id
    else:
        if customer:
            if not idurl:
                idurl = GlobalUserToIDURL(customer)
            if customer.count('$'):
                key_alias, _, _ = customer.rpartition('$')
            if customer.count('!'):
                user_and_key, _, _ = customer.rpartition('@')
                _, _, key_alias = user_and_key.rpartition('!')
        if idurl:
            from bitdust.lib import nameurl
            _, idhost, port, filename = nameurl.UrlParse(idurl)
            if port:
                idhost += '_' + str(port)
            user = filename.strip()[0:-4]
        if key_alias:
            out = output_format.format(user=user, key_alias=key_alias)
        else:
            out = user
        out += '@{}'.format(idhost)
    if path:
        out += ':{}'.format(path)
        if version:
            out += '#{}'.format(version)
    return strng.to_text(out)


def ParseIDURL(idurl):
    return ParseGlobalID(UrlToGlobalID(idurl, include_key=False))


def ParseGlobalID(inp, detect_version=False, as_field=True, fast=True):
    """
    Split input string by parts according to different global ID formats:

    For such input (global resource path):

        "group_abc$alice@first-machine.com:myfiles/animals/cat.png#F20160313043757PM"

    returns such dictionary object:

        {
            "user": "alice",
            "key_alias": "group_abc",
            "key_id": "group_abc$alice@first-machine.com",
            "idhost": "first-machine.com",
            "customer": "alice@first-machine.com",
            "idurl": b"http://first-machine.com/alice.xml",
            "path": "myfiles/animals/cat.png",
            "version": "F20160313043757PM",
        }

    For such input (global path ID) with `detect_version=True`:

        "group_abc$alice@first-machine.com:1/2/3/F20160313043757PM/4-5-Parity"

    returns such dictionary object:

        {
            "user": "alice",
            "key_alias": "group_abc",
            "key_id": "group_abc$alice@first-machine.com",
            "idhost": "first-machine.com",
            "customer": "alice@first-machine.com",
            "idurl": b"http://first-machine.com/alice.xml",
            "path": "1/2/3/F20160313043757PM/4-5-Parity",
            "version": "F20160313043757PM",
        }
    """
    result = {
        'user': '',
        'key_alias': '',
        'key_id': '',
        'idhost': '',
        'customer': '',
        'idurl': b'',
        'path': '',
        'version': '',
    }
    if not inp:
        if as_field:
            from bitdust.userid import id_url
            result['idurl'] = id_url.field(result['idurl'])
        return result
    inp = strng.to_text(inp)
    if inp.count('&') == 2:
        # this is GLOBAL_ID_QUEUE_ID format : just need to get rid of the last supplier_id part and
        # translate it into GLOBAL_ID_KEY_USER format
        inp, _, _ = inp.strip().rpartition('&')
        inp = inp.replace('&', '$')
    if inp.count(':'):
        user, _, path = inp.strip().rpartition(':')
    else:
        if inp.count('@'):
            user = inp
            path = ''
        else:
            user = ''
            path = inp
    if user:
        user_and_key, _, idhost = user.strip().rpartition('@')
        if not user_and_key or not idhost:
            return result
        try:
            if fast:
                _key_alias, _, _user = user_and_key.rpartition('$')
                result['key_alias'] = _key_alias
                result['user'] = _user
            else:
                user_key = re.match(_REGEX_GLOBAL_ID_KEY_USER, user_and_key)
                if not user_key:
                    user_key = re.match(_REGEX_GLOBAL_ID_USER_KEY, user_and_key)
                if user_key:
                    result['user'] = user_key.group('user')
                    result['key_alias'] = user_key.group('key_alias')
                else:
                    result['user'] = user_and_key
        except:
            return result
        result['idhost'] = idhost
        if result['idhost'].count('_'):
            _pos = result['idhost'].rfind('_')
            port = result['idhost'][_pos + 1:]
            try:
                port = int(port)
            except:
                port = -1
            if port >= 0:
                result['idhost'] = '%s:%d' % (result['idhost'][:_pos], port)
        if result['user'] and result['idhost']:
            result['idurl'] = strng.to_bin('http://{}/{}.xml'.format(result['idhost'], result['user']))
            result['customer'] = '{}@{}'.format(result['user'], result['idhost'].replace(':', '_'))
    if path:
        if path.count('#'):
            path, _, version = path.rpartition('#')
            result['version'] = version
        result['path'] = path
        if detect_version:
            try:
                from bitdust.lib import packetid
                backupID, _, fileName = path.rpartition('/')
                if packetid.IsPacketNameCorrect(fileName):
                    _, _, versionName = backupID.rpartition('/')
                    result['version'] = versionName
            except:
                pass
    if not result['key_alias']:
        result['key_alias'] = 'master'
    if result['customer']:
        result['key_id'] = MakeGlobalKeyID(result['key_alias'], result['customer'])
    if as_field:
        from bitdust.userid import id_url
        result['idurl'] = id_url.field(result['idurl'])
    return result


def NormalizeGlobalID(inp, detect_version=False, as_field=True):
    """
    Input `inp` is a string or glob_path_id dict.
    This will fill out missed/empty fields from existing data.
    Such an order:
        1. if no idurl : use my local identity,
        2. if no customer : use idurl
        3. if no user : use customer
        4. if no key alias : use "master"
        5. if no idhost : use idurl
    """
    from bitdust.userid import my_id
    if isinstance(inp, dict):
        g = inp
    else:
        g = ParseGlobalID(inp, detect_version=detect_version)
    if not g['idurl']:
        g['idurl'] = my_id.getIDURL().to_bin()
    if as_field:
        from bitdust.userid import id_url
        g['idurl'] = id_url.field(g['idurl'])
    if not g['customer']:
        g['customer'] = UrlToGlobalID(g['idurl'])
    if not g['user']:
        g['user'] = g['customer'].split('@')[0]
    if not g['key_alias']:
        g['key_alias'] = 'master'
    if not g['idhost']:
        from bitdust.lib import nameurl
        g['idhost'] = nameurl.GetHost(g['idurl'])
    if not g['key_id']:
        g['key_id'] = MakeGlobalKeyID(g['key_alias'], g['customer'])
    return g


def CanonicalID(inp, include_key=True):
    parts = NormalizeGlobalID(ParseGlobalID(inp))
    if include_key:
        parts['key_alias'] = parts.get('key_alias') or 'master'
    else:
        parts['key_alias'] = ''
    return MakeGlobalID(**parts)


def SubstitutePacketID(packet_id, idurl=None, customer=None, key_id=None, path=None, key_alias=None):
    from bitdust.lib import nameurl
    g = ParseGlobalID(packet_id, as_field=False)
    if key_id is not None:
        g['key_id'] = key_id
        key_alias, _, customer = key_id.rpartition('$')
    if key_alias is not None:
        g['key_alias'] = key_alias
    if path is not None:
        g['path'] = path
    if idurl is not None:
        g['idurl'] = idurl
        g['customer'] = UrlToGlobalID(g['idurl'])
        _, idhost, port, filename = nameurl.UrlParse(idurl)
        if port:
            idhost += '_' + str(port)
        g['user'] = filename.strip()[0:-4]
        if key_id is None:
            g['key_id'] = MakeGlobalKeyID(g['key_alias'], g['customer'])
    if customer is not None:
        g['customer'] = customer
        g['idurl'] = glob2idurl(g['customer'], as_field=False)
        _, idhost, port, filename = nameurl.UrlParse(idurl)
        if port:
            idhost += '_' + str(port)
        g['user'] = filename.strip()[0:-4]
        if key_id is None:
            g['key_id'] = MakeGlobalKeyID(g['key_alias'], g['customer'])
    return MakeGlobalID(**g)


#------------------------------------------------------------------------------


def UrlToGlobalID(url, include_key=False):
    if not url:
        return url
    from bitdust.lib import nameurl
    from bitdust.userid import id_url
    if isinstance(url, id_url.ID_URL_FIELD):
        # small optimization
        return url.to_id()
    _, host, port, _, filename = nameurl.UrlParseFast(url)
    if filename.count('.'):
        username = filename.split('.')[0]
    else:
        username = filename
    if port:
        host = '%s_%s' % (host, port)
    if include_key:
        username = 'master$%s' % username
    return strng.to_text('%s@%s' % (username, host))


def GlobalUserToIDURL(inp, as_field=True):
    inp = strng.to_text(inp)
    user, _, idhost = inp.strip().rpartition('@')
    if not user:
        return None
    if not idhost:
        return None
    _, _, user = user.strip().rpartition('$')
    if idhost.count('_'):
        # we can do that because domain names never use "_" symbol
        _pos = idhost.rfind('_')
        port = idhost[_pos + 1:]
        try:
            port = int(port)
        except:
            port = -1
        if port >= 0:
            idhost = '%s:%d' % (idhost[:_pos], port)
    if not as_field:
        return strng.to_bin('http://{}/{}.xml'.format(idhost, user))
    from bitdust.userid import id_url
    return id_url.field('http://{}/{}.xml'.format(idhost, user))


#------------------------------------------------------------------------------


def IsValidGlobalUser(inp):
    if not inp:
        return False
    inp = strng.to_text(inp)
    if inp.count('@') != 1:
        return False
    user, _, idhost = inp.strip().rpartition('@')
    if not user:
        return False
    if not idhost:
        return False
    # TODO: validate user and idhost
    return True


def IsFullGlobalID(inp):
    if not inp:
        return False
    inp = strng.to_text(inp)
    user, _, remote_path = inp.strip().rpartition(':')
    if not IsValidGlobalUser(user):
        return False
    if not remote_path:
        return False
    return True


#------------------------------------------------------------------------------


def MakeGlobalQueueID(queue_alias, owner_id, supplier_id):
    global _FORMAT_GLOBAL_ID_QUEUE_ID
    return _FORMAT_GLOBAL_ID_QUEUE_ID.format(
        queue_alias=strng.to_text(queue_alias),
        owner_id=strng.to_text(owner_id),
        supplier_id=strng.to_text(supplier_id),
    )


def ParseGlobalQueueID(queue_id):
    global _REGEX_GLOBAL_ID_QUEUE_ID
    global _REGEX_OBJ_GLOBAL_ID_QUEUE_ID
    if _REGEX_OBJ_GLOBAL_ID_QUEUE_ID is None:
        _REGEX_OBJ_GLOBAL_ID_QUEUE_ID = re.compile(_REGEX_GLOBAL_ID_QUEUE_ID)
    ret = {
        'queue_alias': '',
        'owner_id': '',
        'supplier_id': '',
    }
    result = _REGEX_OBJ_GLOBAL_ID_QUEUE_ID.match(queue_id)
    if not result:
        return ret
    ret['queue_alias'] = strng.to_text(result.group('queue_alias'))
    ret['owner_id'] = strng.to_text(result.group('owner_id'))
    ret['supplier_id'] = strng.to_text(result.group('supplier_id'))
    return ret


def SplitGlobalQueueID(queue_id, split_queue_alias=True):
    queue_alias_owner_id, _, supplier_id = queue_id.rpartition('&')
    if not split_queue_alias:
        return queue_alias_owner_id, supplier_id
    queue_alias, _, owner_id = queue_alias_owner_id.partition('&')
    return queue_alias, owner_id, supplier_id


def GetGlobalQueueOwnerIDURL(queue_id, as_field=True):
    queue_alias_owner_id, _, _ = queue_id.rpartition('&')
    _, _, owner_id = queue_alias_owner_id.partition('&')
    owner_idurl = glob2idurl(owner_id, as_field=as_field)
    return owner_idurl


def GetGlobalQueueKeyID(queue_id):
    queue_alias_owner_id, _, _ = queue_id.rpartition('&')
    queue_alias, _, owner_id = queue_alias_owner_id.partition('&')
    key_id = MakeGlobalKeyID(queue_alias, owner_id)
    return key_id



#!/usr/bin/python
# service_entangled_dht.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_entangled_dht.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_entangled_dht
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return EntangledDHTService()


class EntangledDHTService(LocalService):

    service_name = 'service_entangled_dht'
    config_path = 'services/entangled-dht/enabled'
    start_suspended = True

    def dependent_on(self):
        return [
            'service_udp_datagrams',
        ]

    def network_configuration(self):
        import re
        from bitdust.main import config
        from bitdust_forks.entangled.kademlia import constants  # @UnresolvedImport
        known_dht_nodes_str = config.conf().getString('services/entangled-dht/known-nodes').strip()
        known_dht_nodes = []
        if known_dht_nodes_str:
            for dht_node_str in re.split('\n|;|,| ', known_dht_nodes_str):
                if dht_node_str.strip():
                    try:
                        dht_node = dht_node_str.strip().split(':')
                        dht_node_host = dht_node[0].strip()
                        dht_node_port = int(dht_node[1].strip())
                    except:
                        continue
                    known_dht_nodes.append({
                        'host': dht_node_host,
                        'udp_port': dht_node_port,
                    })
        if not known_dht_nodes:
            from bitdust.main import network_config
            default_network_config = network_config.read_network_config_file()
            known_dht_nodes = default_network_config['service_entangled_dht']['known_nodes']
        return {
            'bucket_size': constants.k,
            'default_age': constants.dataExpireSecondsDefaut,
            'max_age': constants.dataExpireTimeout,
            'parallel_calls': constants.alpha,
            'refresh_timeout': constants.refreshTimeout,
            'rpc_timeout': constants.rpcTimeout,
            'known_nodes': known_dht_nodes,
        }

    def start(self):
        from twisted.internet.defer import Deferred, succeed
        from bitdust.logs import lg
        from bitdust.dht import dht_records
        from bitdust.dht import dht_service
        from bitdust.dht import known_nodes
        from bitdust.main import settings
        from bitdust.main.config import conf
        conf().addConfigNotifier('services/entangled-dht/udp-port', self._on_udp_port_modified)
        known_seeds = known_nodes.nodes()
        dht_layers = list(dht_records.LAYERS_REGISTRY.keys())
        dht_service.init(
            udp_port=settings.getDHTPort(),
            dht_dir_path=settings.ServiceDir('service_entangled_dht'),
            open_layers=dht_layers,
        )
        lg.info('DHT known seed nodes are : %r   DHT layers are : %r' % (known_seeds, dht_layers))
        self.starting_deferred = Deferred()
        d = dht_service.connect(
            seed_nodes=known_seeds,
            layer_id=0,
            attach=True,
        )
        d.addCallback(self._on_connected)
        d.addErrback(self._on_connect_failed)
        return self.starting_deferred or succeed(True)

    def stop(self):
        from bitdust.dht import dht_records
        from bitdust.dht import dht_service
        from bitdust.main.config import conf
        for layer_id in dht_records.LAYERS_REGISTRY.keys():
            dht_service.close_layer(layer_id)
        dht_service.node().remove_rpc_callback('request')
        dht_service.node().remove_rpc_callback('store')
        conf().removeConfigNotifier('services/entangled-dht/udp-port')
        dht_service.disconnect()
        dht_service.shutdown()
        return True

    def on_suspend(self, *args, **kwargs):
        from bitdust.dht import dht_service
        dht_service.disconnect()
        return True

    def on_resume(self, *args, **kwargs):
        from bitdust.dht import dht_service
        dht_service.reconnect()
        return True

    def health_check(self):
        return True

    def _on_connected(self, ok):
        from twisted.internet.defer import DeferredList
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import known_nodes
        from bitdust.main.config import conf
        from bitdust.services import driver
        lg.info('DHT node connected    ID0=[%s] : %r' % (dht_service.node().layers[0], ok))
        dht_service.node().add_rpc_callback('store', self._on_dht_rpc_store)
        dht_service.node().add_rpc_callback('request', self._on_dht_rpc_request)
        known_seeds = known_nodes.nodes()
        dl = []
        attached_layers = conf().getString('services/entangled-dht/attached-layers', default='')
        if attached_layers:
            attached_layers = list(filter(None, map(lambda v: int(str(v).strip()), attached_layers.split(','))))
        else:
            attached_layers = []
        lg.info('reading attached DHT layers from configuration: %r' % attached_layers)
        all_services_attached_layers = driver.get_attached_dht_layers().values()
        combined_services_attached_layers = set()
        count_combined = len(list(map(combined_services_attached_layers.update, all_services_attached_layers)))
        services_attached_layers = list(combined_services_attached_layers)
        lg.info('combined attached DHT layers from %d services: %r' % (count_combined, services_attached_layers))
        attached_layers = list(set(attached_layers + services_attached_layers))
        lg.info('DHT layers to be attached at startup: %r' % attached_layers)
        for layer_id in attached_layers:
            dl.append(dht_service.open_layer(
                layer_id=layer_id,
                seed_nodes=known_seeds,
                connect_now=True,
                attach=True,
            ))
        if dl:
            d = DeferredList(dl)
            d.addCallback(self._on_layers_attached)
            d.addErrback(self._on_connect_failed)
        else:
            if self.starting_deferred and not self.starting_deferred.called:
                self.starting_deferred.callback(True)
        return ok

    def _on_layers_attached(self, ok):
        if self.starting_deferred and not self.starting_deferred.called:
            self.starting_deferred.callback(True)
        return ok

    def _on_connect_failed(self, err):
        from bitdust.logs import lg
        lg.err('DHT connect failed : %r' % err)
        if self.starting_deferred and not self.starting_deferred.called:
            self.starting_deferred.errback(err)
        return err

    def _on_udp_port_modified(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        from bitdust.logs import lg
        lg.info('DHT udp port modified %s->%s : %s' % (oldvalue, value, path))
        if network_connector.A():
            network_connector.A('reconnect')

    def _on_dht_rpc_store(self, key, value, originalPublisherID, age, expireSeconds, **kwargs):
        from bitdust.dht import dht_service
        return dht_service.validate_before_store(key, value, originalPublisherID, age, expireSeconds, **kwargs)

    def _on_dht_rpc_request(self, key, **kwargs):
        from bitdust.dht import dht_service
        return dht_service.validate_before_request(key, **kwargs)



#!/usr/bin/python
# signed.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (signed.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: signed.

These packets usually hold on the order of 1 MB.
Something equal to a packet number so we can detect duplicates in transport.
Packet Fields are all strings (no integers, objects, etc)
    - Command : Legal Commands are in bitdust/lib/commands.py
    - OwnerID : who owns this data and pays bills - http://cate.com/id1.xml
    - CreatorID : this is signer - http://cate.com/id1.xml - might be an authorized scrubber
    - PacketID : string of the above 4 "Number"s with "-" separator to uniquely identify a packet
                on the local machine.  Can be used for filenames, and to prevent duplicates.
    - Date : create a string to remember current world time
    - Payload : main body of binary data
    - RemoteID : want full IDURL for other party so troublemaker could not
                use his packets to mess up other nodes by sending it to them
    - Signature : signature on Hash is always by CreatorID
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys

from twisted.internet import threads

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.p2p import commands

from bitdust.lib import packetid
from bitdust.lib import nameurl
from bitdust.lib import utime
from bitdust.lib import strng
from bitdust.lib import serialization

from bitdust.contacts import contactsdb

from bitdust.crypt import key

from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------


class Packet(object):
    """
    Init with: Command, OwnerID, CreatorID, PacketID, Payload, RemoteID The
    core class.

    Represents a data packet in the network. Payload can be encrypted
    using bitdust.key.encrypted.Block. We expect remote user run the
    correct software. His BitDust must verify signature of that packet .
    If you want to encrypt the fields and so hide that service traffic
    completely - do that in the transport protocols. Need to transfer
    our public key to remote peer and than he can send us a safe
    messages. This is outside work, here is most important things to
    make all network working.
    """
    def __init__(
        self,
        Command,
        OwnerID,
        CreatorID,
        PacketID,
        Payload,
        RemoteID,
        KeyID=None,
        Date=None,
        Signature=None,
    ):
        """
        Init all fields and sign the packet.
        """
        # Legal Commands are in commands.py
        self.Command = strng.to_text(Command)
        # who owns this data and pays bills - http://somehost.com/id1.xml
        self.OwnerID = id_url.field(OwnerID)
        # signer - http://cate.com/id1.xml - might be an authorized scrubber
        self.CreatorID = id_url.field(CreatorID)
        # string of the above 4 "Number"s with "-" separator to uniquely identify a packet
        # on the local machine.  Can be used for filenames, and to prevent duplicates.
        self.PacketID = strng.to_text(PacketID)
        # create a string to remember current world time
        self.Date = strng.to_text(Date or utime.sec1970_to_datetime_utc().strftime('%Y/%m/%d %I:%M:%S %p'))
        # datetime.datetime.now().strftime("%Y/%m/%d %I:%M:%S %p")
        # main body of binary data
        self.Payload = strng.to_bin(Payload)
        # want full IDURL for other party so troublemaker could not
        # use his packets to mess up other nodes by sending it to them
        self.RemoteID = id_url.field(RemoteID)
        # which private key to use to generate signature
        self.KeyID = strng.to_text(KeyID or my_id.getGlobalID(key_alias='master'))
        if Signature:
            self.Signature = Signature
        else:
            # signature on Hash is always by CreatorID
            self.Signature = None
            # must be signed to be valid
            self.Sign()
        # stores list of related objects packet_in() or packet_out()
        self.Packets = []

    def __repr__(self):
        args = '%s(%s)' % (str(self.Command), str(self.PacketID))
        if _Debug:
            if lg.is_debug(_DebugLevel):
                args += ' %s|%s for %s' % (nameurl.GetName(self.OwnerID), nameurl.GetName(self.CreatorID), nameurl.GetName(self.RemoteID))
        return 'signed{%s}' % args

    def Sign(self):
        """
        Call ``GenerateSignature`` and save the result.
        Usually just done at packet creation.
        """
        self.Signature = self.GenerateSignature()
        return self

    def GenerateHashBase(self):
        """
        This make a long string containing all needed fields of ``packet``
        (without Signature).
        Just to be able to generate a hash of the whole packet .
        """
        sep = b'-'
        stufftosum = b''
        try:
            stufftosum += strng.to_bin(self.Command)
            stufftosum += sep
            stufftosum += self.OwnerID.original()
            stufftosum += sep
            stufftosum += self.CreatorID.original()
            stufftosum += sep
            stufftosum += strng.to_bin(self.PacketID)
            stufftosum += sep
            stufftosum += strng.to_bin(self.Date)
            stufftosum += sep
            stufftosum += strng.to_bin(self.Payload)
            stufftosum += sep
            stufftosum += self.RemoteID.original()
            stufftosum += sep
            stufftosum += strng.to_bin(self.KeyID)
        except Exception as exc:
            lg.exc()
            raise exc
#         if _Debug:
#             if _LogSignVerify:
#                 try:
#                     from bitdust.main import settings
#                     open(os.path.join(settings.LogsDir(), 'crypt.log'), 'wb').write(b'\nGenerateHashBase:\n' + stufftosum + b'\n\n')
#                 except:
#                     lg.exc()
        return stufftosum

    def GenerateHash(self):
        """
        Call ``crypt.key.Hash`` to create a hash code for that ``packet``.
        """
        return key.Hash(self.GenerateHashBase())

    def GenerateSignature(self):
        """
        Call ``crypt.key.Sign`` to generate digital signature.
        """
        _hash_base = self.GenerateHash()
        signature = key.Sign(_hash_base)
        # if not self.KeyID or self.KeyID == my_id.getGlobalID(key_alias='master'):
        #     signature = key.Sign(_hash_base)
        # else:
        #     signature = my_keys.sign(self.KeyID, _hash_base)
        #         if _Debug:
        #             if _LogSignVerify:
        #                 try:
        #                     from bitdust.main import settings
        #                     try:
        #                         from Cryptodome.Util import number
        #                     except:
        #                         from Crypto.Util import number  # @UnresolvedImport @Reimport
        #                     open(os.path.join(settings.LogsDir(), 'crypt.log'), 'wb').write(b'\GenerateSignature:\n' + strng.to_bin(number.long_to_bytes(signature)) + b'\n\n')
        #                 except:
        #                     lg.exc()
        return signature

    def SignatureChecksOut(self, raise_signature_invalid=False):
        """
        This check correctness of signature, uses ``crypt.key.Verify``. To
        verify we need 3 things:

        - the packet ``Creator`` identity ( it keeps the public key ),
        - hash of that packet - just call ``GenerateHash()`` to make it,
        - the signature itself.
        """
        CreatorIdentity = contactsdb.get_contact_identity(self.CreatorID)
        if CreatorIdentity is None:
            # OwnerIdentity = contactsdb.get_contact_identity(self.OwnerID)
            # if OwnerIdentity is None:
            #     lg.err("could not get Identity for %s so returning False" % self.CreatorID.to_text())
            #     return False
            # CreatorIdentity = OwnerIdentity
            if raise_signature_invalid:
                raise Exception('can not verify signed packet, unknown identity %r' % self.CreatorID)
            lg.err('could not get Identity for %r so returning False' % self.CreatorID)
            return False


#         if _Debug:
#             if _LogSignVerify:
#                 try:
#                     from bitdust.main import settings
#                     try:
#                         from Cryptodome.Util import number
#                     except:
#                         from Crypto.Util import number  # @UnresolvedImport @Reimport
#                     open(os.path.join(settings.LogsDir(), 'crypt.log'), 'wb').write(b'\SignatureChecksOut:\n' + strng.to_bin(number.long_to_bytes(self.Signature)) + b'\n\n')
#                 except:
#                     lg.exc()

        Result = key.Verify(CreatorIdentity, self.GenerateHash(), self.Signature)

        #         if _Debug:
        #             if _LogSignVerify:
        #                 try:
        #                     from bitdust.main import settings
        #                     open(os.path.join(settings.LogsDir(), 'crypt.log'), 'wb').write(b'\Result:' + strng.to_bin(str(Result)) + b'\n\n')
        #                 except:
        #                     lg.exc()

        return Result

    def Ready(self):
        """
        I was playing with generating signatures in separate thread, so this is
        just to check that Signature already exists.
        """
        return self.Signature is not None

    def Valid(self, raise_signature_invalid=False):
        """
        ``Valid()`` should check every one of packet header fields: 1) that
        command is one of the legal commands 2) signature is good (which means
        the hashcode is good) Rest PREPRO: 3) all the number fields are just
        numbers 4) length is within legal limits 5) check that URL is a good
        URL 6) that DataOrParity is either "data" or "parity" 7) that Creator
        is equal to owner or a scrubber for owner 8) etc.
        """
        if not self.Ready():
            if _Debug:
                lg.out(_DebugLevel, 'signed.Valid packet is not ready yet ' + str(self))
            return False
        if not commands.IsCommand(self.Command):
            lg.warn('signed.Valid bad Command ' + str(self.Command))
            return False
        if not self.SignatureChecksOut(raise_signature_invalid=raise_signature_invalid):
            if raise_signature_invalid:
                creator_xml = contactsdb.get_contact_identity(self.CreatorID)
                if creator_xml:
                    creator_xml = creator_xml.serialize(as_text=True)
                owner_xml = contactsdb.get_contact_identity(self.OwnerID)
                if owner_xml:
                    owner_xml = owner_xml.serialize(as_text=True)
                raise Exception('signature is not valid for %r:\n\n%r\n\ncreator:\n\n%r\n\nowner:\n\n%r' % (self, self.Serialize(), creator_xml, owner_xml))
            lg.warn('signed.Valid Signature IS NOT VALID!!!')
            return False
        return True

    def BackupID(self):
        """
        """
        backupID, _, _ = self.PacketID.rpartition('/')
        return backupID

    def BlockNumber(self):
        """
        A wrapper for ``lib.packetid.BlockNumber`` on top of self.PacketID.
        """
        return packetid.BlockNumber(self.PacketID)

    def DataOrParity(self):
        """
        A wrapper for ``lib.packetid.DataOrParity`` on top of self.PacketID.
        """
        return packetid.DataOrParity(self.PacketID)

    def SupplierNumber(self):
        """
        A wrapper for ``lib.packetid.SupplierNumber`` on top of self.PacketID.
        """
        return packetid.SupplierNumber(self.PacketID)

    def Serialize(self):
        """
        Create a string from packet object.
        This is useful when need to save the packet on disk or send via network.
        """
        dct = {
            'm': self.Command,
            'o': self.OwnerID.original(),
            'c': self.CreatorID.original(),
            'i': self.PacketID,
            'd': self.Date,
            'p': self.Payload,
            'r': self.RemoteID.original(),
            'k': self.KeyID,
            's': self.Signature,
        }
        src = serialization.DictToBytes(dct, encoding='latin1')
        # if _Debug:
        #     lg.out(_DebugLevel, 'signed.Serialize %d bytes %s(%s) %s/%s/%s KeyID=%s\n%r' % (
        #         len(src), self.Command, self.PacketID, nameurl.GetName(self.OwnerID),
        #         nameurl.GetName(self.CreatorID), nameurl.GetName(self.RemoteID), self.KeyID, dct['s']))
        return src

    def __len__(self):
        """
        Return a length of serialized packet .
        """
        return len(self.Serialize())


class PacketZeroSigned(Packet):
    """
    I was playing with hacking packets and do some debug also.
    """
    def GenerateSignature(self):
        return '0'


def Unserialize(data):
    """
    We expect here a string containing a whole packet object in text form.
    Will return a real object in the memory from given string.
    All class fields are loaded, signature can be verified to be sure - it was truly original string.
    """
    if data is None:
        return None

    dct = serialization.BytesToDict(data, keys_to_text=True, encoding='latin1')

    # if _Debug:
    #     lg.out(_DebugLevel, 'signed.Unserialize %d bytes : %r' % (len(data), dct['s']))

    try:
        Command = strng.to_text(dct['m'])
        OwnerID = dct['o']
        CreatorID = dct['c']
        PacketID = strng.to_text(dct['i'])
        Date = strng.to_text(dct['d'])
        Payload = dct['p']
        RemoteID = dct['r']
        KeyID = strng.to_text(dct['k'])
        Signature = dct['s']
    except:
        lg.exc()
        return None

    try:
        newobject = Packet(
            Command=Command,
            OwnerID=OwnerID,
            CreatorID=CreatorID,
            PacketID=PacketID,
            Date=Date,
            Payload=Payload,
            RemoteID=RemoteID,
            KeyID=KeyID,
            Signature=Signature,
        )
    except:
        if _Debug:
            lg.args(
                _DebugLevel,
                Command=Command,
                OwnerID=OwnerID,
                CreatorID=CreatorID,
                PacketID=PacketID,
                Date=Date,
                Payload=Payload,
                RemoteID=RemoteID,
                KeyID=KeyID,
                Signature=Signature,
            )
        lg.exc()
        return None

    # if _Debug:
    #     lg.args(_DebugLevel, Command=Command, PacketID=PacketID, OwnerID=OwnerID, CreatorID=CreatorID, RemoteID=RemoteID)

    return newobject


def MakePacket(Command, OwnerID, CreatorID, PacketID, Payload, RemoteID):
    """
    Just calls the constructor of packet class.
    """
    result = Packet(Command, OwnerID, CreatorID, PacketID, Payload, RemoteID)
    return result


def MakePacketInThread(CallBackFunc, Command, OwnerID, CreatorID, PacketID, Payload, RemoteID):
    """
    Signing packets is not atomic operation, so can be moved out from the main
    thread.
    """
    d = threads.deferToThread(MakePacket, Command, OwnerID, CreatorID, PacketID, Payload, RemoteID)
    d.addCallback(CallBackFunc)


def MakePacketDeferred(Command, OwnerID, CreatorID, PacketID, Payload, RemoteID):
    """
    Another nice way to create a signed packet .
    """
    return threads.deferToThread(MakePacket, Command, OwnerID, CreatorID, PacketID, Payload, RemoteID)


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(18)
    from bitdust.main import settings
    settings.init()
    key.InitMyKey()
    from bitdust.userid import identity
    from bitdust.contacts import identitycache
    if len(sys.argv) > 2:
        creator_ident = identity.identity(xmlsrc=bpio.ReadTextFile(sys.argv[2]))
        identitycache.UpdateAfterChecking(idurl=creator_ident.getIDURL(), xml_src=creator_ident.serialize())
    if len(sys.argv) > 3:
        owner_ident = identity.identity(xmlsrc=bpio.ReadTextFile(sys.argv[3]))
        identitycache.UpdateAfterChecking(idurl=owner_ident.getIDURL(), xml_src=owner_ident.serialize())
    p = Unserialize(bpio.ReadBinaryFile(sys.argv[1]))
    print(p.Valid())
    print(p)
    settings.shutdown()



#!/usr/bin/python
# service_http_connections.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_http_connections.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_http_connections
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return HTTPConnectionsService()


class HTTPConnectionsService(LocalService):

    service_name = 'service_http_connections'
    config_path = 'services/http-connections/enabled'

    def dependent_on(self):
        return [
            'service_network',
        ]

    def start(self):
        from bitdust.main.config import conf
        conf().addConfigNotifier('services/http-connections/http-port', self._on_http_port_modified)
        return True

    def stop(self):
        from bitdust.main.config import conf
        conf().removeConfigNotifier('services/http-connections/http-port')
        return True

    def _on_http_port_modified(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        from bitdust.logs import lg
        lg.out(2, 'service_http_connections._on_http_port_modified : %s->%s : %s' % (oldvalue, value, path))
        network_connector.A('reconnect')



#!/usr/bin/python
# udp.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from io import BytesIO

#------------------------------------------------------------------------------

import sys
import time

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet import protocol
from twisted.internet import task
from twisted.internet.defer import DeferredList

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

from bitdust.system import bpio

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

_Listeners = {}
_LastDatagramReceivedTime = 0

#------------------------------------------------------------------------------

CMD_PING = b'p'
CMD_GREETING = b'g'
CMD_DATA = b'd'
CMD_ACK = b'k'
CMD_ALIVE = b'a'
CMD_STUN = b's'
CMD_MYIPPORT = b'm'

#------------------------------------------------------------------------------


def listen(port, proto=None):
    if port in list(listeners().keys()):
        lg.warn('already started on port %d' % port)
        if _Debug:
            lg.out(6, '            %s' % str(list(listeners().keys())))
        return listeners()[port]
    if proto is None:
        listeners()[port] = reactor.listenUDP(port, CommandsProtocol())  # @UndefinedVariable
    else:
        listeners()[port] = reactor.listenUDP(port, proto)  # @UndefinedVariable
    listeners()[port].port = port
    if _Debug:
        lg.out(6, 'udp.listen on port %d started' % port)
    return listeners()[port]


def port_closed(x):
    if _Debug:
        lg.out(6, 'udp.port_closed   listeners: %d' % (len(listeners())))
    return x


def close(port):
    if _Debug:
        lg.out(6, 'udp.close  %r' % port)
    l = listeners().pop(port)
    l.protocol.disconnect()
    d = l.stopListening()
    del l
    l = None
    if d:
        d.addCallback(port_closed)
    if _Debug:
        lg.out(6, 'udp.close  STOP listener on UDP port %d' % port)
    return d


def close_all():
    shutlist = []
    l = list(listeners().keys())
    for port in l:
        d = close(port)
        if d:
            shutlist.append(d)
    if _Debug:
        lg.out(6, 'udp.close_all  %d UDP listeners were closed' % len(shutlist))
    return DeferredList(shutlist)


#------------------------------------------------------------------------------


def listeners():
    global _Listeners
    return _Listeners


def proto(port):
    if port not in listeners():
        return None
    return listeners()[port].protocol


def listener(port):
    if port not in listeners():
        return None
    return listeners()[port]


#------------------------------------------------------------------------------


def send_command(from_port, command, data, address):
    p = proto(from_port)
    if not p:
        lg.warn('port %d is not opened to listen' % from_port)
        return False
    result = p.sendCommand(command, data, address)
    p = None
    return result


def get_last_datagram_time():
    global _LastDatagramReceivedTime
    return _LastDatagramReceivedTime


#------------------------------------------------------------------------------


class BasicProtocol(protocol.DatagramProtocol):
    def __init__(self):
        self.port = None
        self.callbacks = []
        self.stopping = False
        self.bytes_in = 0
        self.bytes_out = 0

    def __del__(self):
        """
        """

    def insert_callback(self, index, cb):
        self.callbacks.insert(index, cb)

    def add_callback(self, cb):
        if cb in self.callbacks:
            lg.warn('Callback method %s already registered' % cb)
        else:
            self.callbacks.append(cb)

    def remove_callback(self, cb):
        if cb not in self.callbacks:
            lg.warn('Callback method %s not registered' % cb)
        else:
            self.callbacks.remove(cb)

    def run_callbacks(self, data, address):
        for cb in self.callbacks:
            if cb(data, address):
                break

    def datagramReceived(self, datagram, address):
        self.bytes_in += len(datagram)
        self.run_callbacks(datagram, address)

    def sendDatagram(self, datagram, address):
        """
        Sends UDP datagram to transport.
        """
        if self.stopping:
            return False
        try:
            self.transport.write(datagram, address)
        except:
            return False
        self.bytes_out += len(datagram)
        return True

    def startProtocol(self):
        """
        """
        if _Debug:
            lg.out(6, 'udp.startProtocol %r' % self)

    def stopProtocol(self):
        """
        """
        if _Debug:
            lg.out(6, 'udp.stopProtocol %r' % self)
        self.port = None
        self.callbacks = []

    def disconnect(self):
        """
        """
        self.stopping = True
        self.callbacks = []


#------------------------------------------------------------------------------


class CommandsProtocol(BasicProtocol):
    """
    Datagram format is::

        | Software | Command ID | Payload |
        | version  |            |         |
        | (1 byte) | (1 byte)   |         |

    Commands have different payload format, see in the code.
    List of valid commands (by ID):

        * 'p' = ``PING``        an empty packet to establish connection
        * 'g' = ``GREETING``    need to give a response when received a ``PING`` packet,
                                payload should contain a global ID of responding user
                                so remote peer can identify who is this.
        * 'd' = ``DATA``        a data packet, payload format will be described bellow.
        * 'r' = ``REPORT``      a response after receiving a ``DATA`` packet,
                                so sender can send next packets.
        * 'a' = ``ALIVE``       periodically need to send an empty packet to keep session alive.
        * 's' = ``STUN``        request remote peer for my external IP:PORT.
        * 'm' = ``MYIPPORT``    response to ``STUN`` packet, payload will contain IP:PORT of remote peer
    """

    SoftwareVersion = b'1'

    def __init__(self):
        self.command_filter_callback = None
        BasicProtocol.__init__(self)

    def set_command_filter_callback(self, cb):
        self.command_filter_callback = cb

    def datagramReceived(self, datagram, address):
        global _LastDatagramReceivedTime
        _LastDatagramReceivedTime = time.time()
        inp = BytesIO(datagram)
        try:
            # version = datagram[0]
            # command = datagram[1]
            # payload = datagram[2:]
            # payloadsz = len(payload)
            datagramsz = len(datagram)
            version = inp.read(1)
            command = inp.read(1)
        except:
            inp.close()
            lg.exc()
            return
        if version != self.SoftwareVersion:
            inp.close()
            lg.warn('different software version: %s' % version)
            return
        if _Debug:
            lg.out(_DebugLevel, '<<< [%s] (%d bytes) from %s, total %d bytes received' % (command, datagramsz, str(address), self.bytes_in))
        # self.bytes_in += datagramsz
        handled = False
        try:
            if self.command_filter_callback:
                handled = self.command_filter_callback(command, datagram, inp, address)
        except:
            lg.exc()
        payload = inp.read()
        inp.close()
        if not handled:
            self.run_callbacks((command, payload), address)
        self.bytes_in += datagramsz

    def sendCommand(self, command, data, address):
        payloadsz = len(data)
        outp = BytesIO()
        try:
            outp.write(self.SoftwareVersion)
            outp.write(strng.to_bin(command))
            outp.write(strng.to_bin(data))
            if _Debug:
                lg.out(_DebugLevel, '>>> [%s] (%d bytes) to %s, total %d bytes sent' % (command, payloadsz + 2, address, self.bytes_out))
            result = self.sendDatagram(outp.getvalue(), address)
        except:
            outp.close()
            lg.exc()
            return None
        outp.close()
        self.bytes_out += payloadsz + 2
        return result


#------------------------------------------------------------------------------


def main():
    bpio.init()
    lg.set_debug_level(18)
    listnport = int(sys.argv[1])

    def received(dgrm, addr):
        send_command(listnport, CMD_ALIVE, 'ok', addr)

    def go(x, port):
        print('go', x)
        l = listen(port)
        l.protocol.add_callback(received)

    def restart(port):
        print('restart')
        if listener(port):
            close(port).addCallback(go, port)
        else:
            go(None, port)

    def ping(fromport, toaddr):
        print('ping')
        send_command(fromport, CMD_PING, b'ping', toaddr)

    if len(sys.argv) > 2:
        addr = sys.argv[2].split(':')
        addr = (addr[0], int(addr[1]))
        listen(listnport)
        task.LoopingCall(ping, listnport, addr).start(1, False)

    else:
        restart(listnport)
        # task.LoopingCall(restart, listnport).start(5)

    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from bitdust_forks.CodernityDB.database import Database
import shutil
import os


def migrate(source, destination):
    """
    Very basic for now
    """
    dbs = Database(source)
    dbt = Database(destination)
    dbs.open()
    dbt.create()
    dbt.close()
    for curr in os.listdir(os.path.join(dbs.path, '_indexes')):
        if curr != '00id.py':
            shutil.copyfile(os.path.join(dbs.path, '_indexes', curr),
                            os.path.join(dbt.path, '_indexes', curr))
    dbt.open()
    for c in dbs.all('id'):
        del c['_rev']
        dbt.insert(c)
    return True


if __name__ == '__main__':
    import sys
    migrate(sys.argv[1], sys.argv[2])



#!/usr/bin/python
# service_identity_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_identity_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_identity_server
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return IdentityServerService()


class IdentityServerService(LocalService):

    service_name = 'service_identity_server'
    config_path = 'services/identity-server/enabled'

    def init(self):
        self.log_events = True

    def dependent_on(self):
        return [
            'service_tcp_connections',
        ]

    def installed(self):
        return True

    def enabled(self):
        from bitdust.main import settings
        return settings.enableIdServer()

    def start(self):
        from bitdust.userid import id_server
        from bitdust.main import settings
        id_server.A('init', (
            settings.getIdServerWebPort(),
            settings.getIdServerTCPPort(),
        ))
        id_server.A('start')
        return True

    def stop(self):
        from bitdust.userid import id_server
        id_server.A('stop')
        id_server.A('shutdown')
        return True



#!/usr/bin/python
# service_tcp_transport.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_tcp_transport.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_tcp_transport
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return TCPTransportService()


class TCPTransportService(LocalService):

    service_name = 'service_tcp_transport'
    config_path = 'services/tcp-transport/enabled'
    proto = 'tcp'
    stop_when_failed = True

    def dependent_on(self):
        return [
            'service_tcp_connections',
            'service_gateway',
        ]

    def start(self):
        from twisted.internet import reactor  # @UnresolvedImport
        from twisted.internet.defer import Deferred
        from bitdust.transport.tcp import tcp_interface
        from bitdust.transport import network_transport
        from bitdust.transport import gateway
        from bitdust.main.config import conf
        self.starting_deferred = Deferred()
        self.interface = tcp_interface.GateInterface()
        self.transport = network_transport.NetworkTransport('tcp', self.interface)
        self.transport.automat('init', (gateway.listener(), self._on_transport_state_changed))
        reactor.callLater(0, self.transport.automat, 'start')  # @UndefinedVariable
        conf().addConfigNotifier('services/tcp-transport/enabled', self._on_enabled_disabled)
        conf().addConfigNotifier('services/tcp-transport/receiving-enabled', self._on_receiving_enabled_disabled)
        return self.starting_deferred

    def stop(self):
        from bitdust.main.config import conf
        conf().removeConfigNotifier('services/tcp-transport/enabled')
        conf().removeConfigNotifier('services/tcp-transport/receiving-enabled')
        t = self.transport
        self.transport = None
        t.automat('shutdown')
        return True

    def installed(self):
        from bitdust.logs import lg
        try:
            from bitdust.transport.tcp import tcp_interface
        except:
            lg.exc()
            return False
        return True

    def _on_transport_state_changed(self, transport, oldstate, newstate):
        from bitdust.logs import lg
        lg.info('%s -> %s in %r  starting_deferred=%r' % (oldstate, newstate, transport, bool(self.starting_deferred)))
        if self.starting_deferred:
            if newstate in [
                'LISTENING',
            ] and oldstate != newstate:
                self.starting_deferred.callback(True)
                self.starting_deferred = None
            elif newstate in [
                'OFFLINE',
            ] and oldstate != newstate and oldstate not in [
                'INIT',
            ]:
                self.starting_deferred.errback(Exception(newstate))
                self.starting_deferred = None


#        if self.transport:
#            from bitdust.p2p import network_connector
#            network_connector.A('network-transport-state-changed', self.transport)

    def _on_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        from bitdust.logs import lg
        lg.out(2, 'service_tcp_transport._on_enabled_disabled : %s->%s : %s' % (oldvalue, value, path))
        network_connector.A('reconnect')

    def _on_receiving_enabled_disabled(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        from bitdust.logs import lg
        lg.out(2, 'service_tcp_transport._on_receiving_enabled_disabled : %s->%s : %s' % (oldvalue, value, path))
        network_connector.A('reconnect')



#!/usr/bin/env python
# message_producer.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (message_producer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: message_producer
.. role:: red

BitDust message_producer() Automat

EVENTS:
    * :red:`broker-connected`
    * :red:`broker-failed`
    * :red:`brokers-found`
    * :red:`brokers-not-found`
    * :red:`connect`
    * :red:`disconnect`
    * :red:`init`
    * :red:`push-message`
    * :red:`shutdown`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import packetid
from bitdust.lib import serialization
from bitdust.lib import utime

from bitdust.main import config

from bitdust.contacts import identitycache

from bitdust.dht import dht_relations

from bitdust.access import groups

from bitdust.p2p import p2p_service_seeker

from bitdust.stream import message

from bitdust.crypt import my_keys

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ActiveMessageProducers = {}
_ActiveMessageProducersByIDURL = {}

#------------------------------------------------------------------------------


def register_message_producer(A):
    global _ActiveMessageProducers
    global _ActiveMessageProducersByIDURL
    if _Debug:
        lg.args(_DebugLevel, instance=repr(A))
    if A.group_key_id in _ActiveMessageProducers:
        raise Exception('message_producer already exist')
    _ActiveMessageProducers[A.group_key_id] = A
    if id_url.is_not_in(A.group_creator_idurl, _ActiveMessageProducersByIDURL):
        _ActiveMessageProducersByIDURL[A.group_creator_idurl] = []
    _ActiveMessageProducersByIDURL[A.group_creator_idurl].append(A)


def unregister_message_producer(A):
    global _ActiveMessageProducers
    global _ActiveMessageProducersByIDURL
    if _Debug:
        lg.args(_DebugLevel, instance=repr(A))
    if id_url.is_not_in(A.group_creator_idurl, _ActiveMessageProducersByIDURL):
        lg.warn('for given customer idurl %r did not found active message producer' % A.group_creator_idurl)
    else:
        if A in _ActiveMessageProducersByIDURL[A.group_creator_idurl]:
            _ActiveMessageProducersByIDURL[A.group_creator_idurl].remove(A)
        else:
            lg.warn('message_producer() instance not found for customer %r' % A.group_creator_idurl)
    _ActiveMessageProducers.pop(A.group_key_id, None)


#------------------------------------------------------------------------------


def list_active_message_producers():
    global _ActiveMessageProducers
    return list(_ActiveMessageProducers.keys())


def get_active_message_producer(group_key_id):
    global _ActiveMessageProducers
    if group_key_id not in _ActiveMessageProducers:
        return None
    return _ActiveMessageProducers[group_key_id]


def find_active_message_producers(group_creator_idurl):
    global _ActiveMessageProducersByIDURL
    result = []
    for automat_index in _ActiveMessageProducersByIDURL.values():
        A = automat.by_index(automat_index)
        if not A:
            continue
        if A.group_creator_idurl == group_creator_idurl:
            result.append(A)
    return result


#------------------------------------------------------------------------------


def start_message_producers():
    started = 0
    for key_id in my_keys.known_keys().keys():
        # if not key_id.startswith('person$'):
        #     continue
        group_key_id = key_id
        existing_message_producer = get_active_message_producer(group_key_id)
        if not existing_message_producer:
            existing_message_producer = MessageProducer(group_key_id)
            existing_message_producer.automat('init')
        if existing_message_producer.state in [
            'DHT_READ?',
            'BROKER?',
            'CONNECTED',
        ]:
            continue
        existing_message_producer.automat('connect')
        started += 1
    return started


def shutdown_message_producers():
    global _ActiveMessageProducers
    stopped = 0
    for group_key_id in list(_ActiveMessageProducers.keys()):
        existing_message_producer = get_active_message_producer(group_key_id)
        if not existing_message_producer:
            continue
        existing_message_producer.automat('shutdown')
        stopped += 1
    return stopped


#------------------------------------------------------------------------------


def do_send_message(active_message_producer, data, result_defer):
    if _Debug:
        lg.args(_DebugLevel, active_message_producer=active_message_producer)
    active_message_producer.automat('push-message', json_payload=data)
    result_defer.callback(True)
    return None


def do_start_message_producer(group_key_id, data, result_defer):
    active_message_producer = get_active_message_producer(group_key_id)
    if _Debug:
        lg.args(_DebugLevel, active_message_producer=active_message_producer)
    if not active_message_producer:
        active_message_producer = MessageProducer(group_key_id)
        active_message_producer.automat('init')
    if active_message_producer.state == 'CONNECTED':
        do_send_message(active_message_producer, data, result_defer)
        return active_message_producer

    def _on_message_producer_state_changed(oldstate, newstate, event_string, args):
        if _Debug:
            lg.args(_DebugLevel, oldstate=oldstate, newstate=newstate, event_string=event_string)
        if newstate == 'CONNECTED' and oldstate != newstate:
            active_message_producer.removeStateChangedCallback(_on_message_producer_state_changed)
            do_send_message(active_message_producer, data, result_defer)
        if newstate == 'DISCONNECTED' and oldstate != newstate:
            active_message_producer.removeStateChangedCallback(_on_message_producer_state_changed)
            if not result_defer.called:
                result_defer.errback(Exception('disconnected'))
        return None

    active_message_producer.addStateChangedCallback(_on_message_producer_state_changed)
    active_message_producer.automat('connect')
    return active_message_producer


def push_message(group_key_id, data):
    creator_idurl = my_keys.get_creator_idurl(group_key_id, as_field=False)
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, creator_idurl=creator_idurl)
    ret = Deferred()
    if not id_url.is_cached(creator_idurl):
        d = identitycache.immediatelyCaching(creator_idurl)
        d.addErrback(ret.errback)
        d.addCallback(lambda *args: do_start_message_producer(group_key_id, data, ret))
        return ret
    do_start_message_producer(group_key_id, data, ret)
    return ret


#------------------------------------------------------------------------------


class MessageProducer(automat.Automat):
    """
    This class implements all the functionality of ``message_producer()`` state machine.
    """
    def __init__(self, group_key_id, debug_level=_Debug, log_events=_Debug, log_transitions=_DebugLevel, **kwargs):
        """
        Builds `message_producer()` state machine.
        """
        self.producer_id = my_id.getGlobalID(key_alias='person')
        self.group_key_id = group_key_id
        self.group_glob_id = global_id.ParseGlobalID(self.group_key_id)
        self.group_creator_id = self.group_glob_id['customer']
        self.group_creator_idurl = self.group_glob_id['idurl']
        self.active_broker_id = None
        super(MessageProducer, self).__init__(name='message_producer_%s' % self.group_creator_id, state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions, publish_events=False, **kwargs)

    def to_json(self):
        j = super().to_json()
        j.update({
            'group_key_id': self.group_key_id,
            'label': my_keys.get_label(self.group_key_id) or '',
            'creator': self.group_creator_idurl,
            'active_broker_id': self.active_broker_id,
        })
        return j

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `message_producer()` machine.
        """

    def register(self):
        """
        """
        automat_index = automat.Automat.register(self)
        register_message_producer(self)
        return automat_index

    def unregister(self):
        """
        """
        unregister_message_producer(self)
        return automat.Automat.unregister(self)

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `message_producer()` state were changed.
        """

    def state_not_changed(self, curstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when some event was fired in the `message_producer()`
        but automat state was not changed.
        """

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'DISCONNECTED'
                self.doInit(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'disconnect' or event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'connect':
                self.state = 'DHT_READ?'
                self.doDHTReadBrokers(*args, **kwargs)
        #---DHT_READ?---
        elif self.state == 'DHT_READ?':
            if event == 'disconnect' or event == 'brokers-not-found':
                self.state = 'DISCONNECTED'
                self.doReportDisconnected(event, *args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'brokers-found':
                self.state = 'BROKER?'
                self.doConnectFirstBroker(*args, **kwargs)
        #---BROKER?---
        elif self.state == 'BROKER?':
            if event == 'disconnect' or event == 'broker-failed':
                self.state = 'DISCONNECTED'
                self.doReportDisconnected(event, *args, **kwargs)
            elif event == 'broker-connected':
                self.state = 'CONNECTED'
                self.doRememberBroker(*args, **kwargs)
                self.doReportConnected(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect':
                self.state = 'DISCONNECTED'
                self.doReportDisconnected(event, *args, **kwargs)
            elif event == 'push-message':
                self.doSendMessageToBroker(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """

    def doDHTReadBrokers(self, *args, **kwargs):
        """
        Action method.
        """
        result = dht_relations.read_customer_message_brokers(
            self.group_creator_idurl,
            positions=list(range(groups.REQUIRED_BROKERS_COUNT)),
            use_cache=False,
        )
        # TODO: add more validations of dht_result
        result.addCallback(self._on_read_customer_message_brokers)
        result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='message_producer.doDHTReadBrokers')
        result.addErrback(lambda err: self.automat('dht-read-failed', err))

    def doConnectFirstBroker(self, *args, **kwargs):
        """
        Action method.
        """
        existing_brokers = args[0]
        if _Debug:
            lg.args(_DebugLevel, existing_brokers=existing_brokers)
        known_brokers = [
            None,
        ]*groups.REQUIRED_BROKERS_COUNT
        top_broker_pos = None
        top_broker_idurl = None
        for broker_pos in range(groups.REQUIRED_BROKERS_COUNT):
            broker_at_position = None
            for existing_broker in existing_brokers:
                if existing_broker['position'] == broker_pos and existing_broker['broker_idurl']:
                    broker_at_position = existing_broker
                    break
            if not broker_at_position:
                continue
            try:
                broker_idurl = broker_at_position['broker_idurl']
            except IndexError:
                broker_idurl = None
            if not broker_idurl:
                lg.warn('broker is empty for %r at position %d' % (self.group_key_id, broker_pos))
                continue
            known_brokers[broker_pos] = broker_idurl
            if _Debug:
                lg.dbg(_DebugLevel, 'found broker %r at position %r for %r' % (broker_idurl, broker_pos, self.group_key_id))
            if top_broker_pos is None:
                top_broker_pos = broker_pos
                top_broker_idurl = broker_idurl
            if broker_pos < top_broker_pos:
                top_broker_pos = broker_pos
                top_broker_idurl = broker_idurl
        if _Debug:
            lg.args(_DebugLevel, known_brokers=known_brokers)
        if top_broker_idurl is None:
            lg.info('did not found any existing brokers for %r' % self.group_key_id)
            self.automat('broker-failed')
            return
        if _Debug:
            lg.args(_DebugLevel, top_broker_pos=top_broker_pos, top_broker_idurl=top_broker_idurl)
        result = p2p_service_seeker.connect_known_node(
            remote_idurl=top_broker_idurl,
            service_name='service_message_broker',
            service_params=lambda idurl: self._do_prepare_service_request_params(idurl, top_broker_pos),
            attempts=1,
        )
        result.addCallback(self._on_broker_connected, top_broker_pos)
        result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='message_producer.doConnectFirstBroker')
        result.addErrback(self._on_message_broker_connect_failed, top_broker_pos)

    def doRememberBroker(self, *args, **kwargs):
        """
        Action method.
        """
        self.active_broker_id = args[0]
        self.active_queue_id = global_id.MakeGlobalQueueID(
            queue_alias=self.group_glob_id['key_alias'],
            owner_id=self.group_creator_id,
            supplier_id=self.active_broker_id,
        )

    def doReportConnected(self, *args, **kwargs):
        """
        Action method.
        """

    def doReportDisconnected(self, event, *args, **kwargs):
        """
        Action method.
        """

    def doSendMessageToBroker(self, *args, **kwargs):
        """
        Action method.
        """
        data = kwargs['json_payload']
        data['msg_type'] = 'personal_message'
        data['action'] = 'read'
        self._do_send_message_to_broker(
            json_payload=data,
            outgoing_counter=None,
            packet_id='personal_%s' % packetid.UniqueID(),
        )

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()
        self.producer_id = None
        self.group_key_id = None
        self.group_glob_id = None
        self.group_creator_id = None
        self.group_creator_idurl = None
        self.active_broker_id = None

    def _on_read_customer_message_brokers(self, brokers_info_list):
        if _Debug:
            lg.args(_DebugLevel, brokers=len(brokers_info_list))
        if not brokers_info_list:
            self.automat('brokers-not-found', [])
            return
        self.automat('brokers-found', brokers_info_list)

    def _on_message_broker_connect_failed(self, err, broker_pos):
        if _Debug:
            lg.args(_DebugLevel, err=err, broker_pos=broker_pos)
        self.automat('broker-failed')

    def _on_broker_connected(self, response_info, broker_pos, *a, **kw):
        if _Debug:
            lg.args(_DebugLevel, resp=response_info, broker_pos=broker_pos)
        self.automat('broker-connected', response_info[0].CreatorID)

    def _do_send_message_to_broker(self, json_payload=None, packet_id=None):
        if packet_id is None:
            packet_id = packetid.UniqueID()
        if _Debug:
            lg.args(_DebugLevel, json_payload=json_payload, packet_id=packet_id)
        raw_payload = serialization.DictToBytes(
            json_payload,
            pack_types=True,
            encoding='utf-8',
        )
        try:
            private_message_object = message.GroupMessage(
                recipient=self.group_key_id,
                sender=self.producer_id,
            )
            private_message_object.encrypt(raw_payload)
        except:
            lg.exc()
            raise Exception('message encryption failed')
        encrypted_payload = private_message_object.serialize()
        d = message.send_message(
            json_data={
                'msg_type': 'queue_message',
                'action': 'produce',
                'created': utime.utcnow_to_sec1970(),
                'payload': encrypted_payload,
                'queue_id': self.active_queue_id,
                'producer_id': self.producer_id,
            },
            recipient_global_id=self.active_broker_id,
            packet_id=packetid.MakeQueueMessagePacketID(self.active_queue_id, packet_id),
            message_ack_timeout=config.conf().getInt('services/private-groups/message-ack-timeout'),
            skip_handshake=True,
            fire_callbacks=False,
        )
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='message_producer._do_send_message_to_broker')
        # d.addCallback(self._on_message_to_broker_sent, packet_id)
        # d.addErrback(self._on_message_to_broker_failed, packet_id)
